"""
Unary and Binary predicates on bounding boxes. Assumes objects with structure
{'x1': ..., 'y1': ..., 'x2': ..., 'y2': ...}.

Unary predicates:
    position(x1, y1, x2, y2, epsilon=0.1) - bbox has positions x1, y1, x2, y2,
        +/- epsilon
    has_value(value_name, value, epsilon=0.1) - value_name (one of x1, y1, x2,
        y2) has value value, +/- epsilon
    area_exactly(area, epsilon=0.1) - area of the bbox is some area
    area_at_least(area) - area is at least some area
    area_at_most(area) - area is at most some area
    area_between(area1, area2) - area is between two values
    width_exactly(width, epsilon=0.1) - width of the bbox is some width
    width_at_least(width) - width is at least some width
    width_at_most(width) - width is at most some width
    width_between(width1, width2) - width is between two values
    height_exactly(height, epsilon=0.1) - height of the bbox is some height
    height_at_least(height) - height is at least some height
    height_at_most(height) - height is at most some height
    height_between(height1, height2) - height is between two values

Binary predicates:
    left_of() - start's x2 is less than end's x1
    right_of() - start's x1 is greater than end's x2
    above() - start's y2 is less than end's y1
    below() - start's y1 is greater than end's y2
    same_area(epsilon=0.1) - start and end have same area
    more_area() - start's area greater than end's area
    less_area() - start's area less than end's area
    same_width(epsilon=0.1) - start and end have same width
    more_width() - start's width greater than end's width
    less_width() - start's width less than end's width
    same_height(epsilon=0.1) - start and end have same height
    more_height() - start's height greater than end's height
    less_height() - start's height less than end's height
    same_value(value_name, epsilon=0.1) - start and end have the same value
        (one of x1, y1, x2, y2), +/- epsilon
"""

'''Unary predicates'''
def _area(bbox):
    return (bbox['x2'] - bbox['x1']) * (bbox['y2'] - bbox['y1'])
def _width(bbox):
    return bbox['x2'] - bbox['x1']
def _height(bbox):
    return bbox['y2'] - bbox['y1']
def position(x1, y1, x2, y2, epsilon=0.1):
    return lambda bbox: (abs(bbox['x1'] - x1) < epsilon and
            abs(bbox['y1'] - y1) < epsilon and
            abs(bbox['x2'] - x2) < epsilon and
            abs(bbox['y2'] - y2) < epsilon)
def has_value(value_name, value, epsilon=0.1):
    return lambda bbox: abs(bbox[value_name] - value) < epsilon
def area_exactly(area, epsilon=0.1):
    return lambda bbox: abs(_area(bbox) - area) < epsilon 
def area_at_least(area):
    return lambda bbox: _area(bbox) >= area
def area_at_most(area):
    return lambda bbox: _area(bbox) <= area
def area_between(area1, area2):
    return lambda bbox: _area(bbox) >= area1 and _area(bbox) <= area2
def width_exactly(width, epsilon=0.1):
    return lambda bbox: abs(_width(bbox) - width) < epsilon 
def width_at_least(width):
    return lambda bbox: _width(bbox) >= width
def width_at_most(width):
    return lambda bbox: _width(bbox) <= width
def width_between(width1, width2):
    return lambda bbox: _width(bbox) >= width1 and _width(bbox) <= width2
def height_exactly(height, epsilon=0.1):
    return lambda bbox: abs(_height(bbox) - height) < epsilon 
def height_at_least(height):
    return lambda bbox: _height(bbox) >= height
def height_at_most(height):
    return lambda bbox: _height(bbox) <= height
def height_between(height1, height2):
    return lambda bbox: _height(bbox) >= height1 and _height(bbox) <= height2
def and_pred_node(pred1, pred2):
    return lambda bbox: pred1(bbox) and pred2(bbox)
def or_pred_node(pred1, pred2):
    return lambda bbox: pred1(bbox) or pred2(bbox)

'''Binary predicates'''
def left_of():
    return lambda bbox1, bbox2: bbox1['x2'] < bbox2['x1']
def right_of():
    return lambda bbox1, bbox2: bbox1['x1'] > bbox2['x2']
def above():
    return lambda bbox1, bbox2: bbox1['y2'] < bbox2['y1']
def below():
    return lambda bbox1, bbox2: bbox1['y1'] > bbox2['y2']
def same_area(epsilon=0.1):
    return lambda bbox1, bbox2: abs(_area(bbox1) - _area(bbox2)) < epsilon
def more_area():
    return lambda bbox1, bbox2: _area(bbox1) > _area(bbox2)
def less_area():
    return lambda bbox1, bbox2: _area(bbox1) < _area(bbox2)
def same_width(epsilon=0.1):
    return lambda bbox1, bbox2: abs(_width(bbox1) - _width(bbox2)) < epsilon
def more_width():
    return lambda bbox1, bbox2: _width(bbox1) > _width(bbox2)
def less_width():
    return lambda bbox1, bbox2: _width(bbox1) < _width(bbox2)
def same_height(epsilon=0.1):
    return lambda bbox1, bbox2: abs(_height(bbox1) - _height(bbox2)) < epsilon
def more_height():
    return lambda bbox1, bbox2: _height(bbox1) > _height(bbox2)
def less_height():
    return lambda bbox1, bbox2: _height(bbox1) < _height(bbox2)
def same_value(value_name, epsilon=0.1):
    return lambda bbox1, bbox2: abs(bbox1[value_name] - bbox2[value_name]) < epsilon



"""Payload Parsers"""

from functools import reduce

def dict_payload_parser(accessor, fields):
    """Parse an object to generate a payload.

    @accesor takes in object and a field name and gets a value.
    @fields is a dict mapping from field names in the payload to field names in
    the object.
    """
    def parse_object(obj):
        return { field_name: accessor(obj, fields[field_name])
                for field_name in fields }

    return parse_object

def label_payload_parser(accessor, label):
    """ Parser that produces { "label": ... } """
    return dict_payload_parser(accessor, { "label": label })

def bbox_payload_parser(accessor, x1="bbox_x1", y1="bbox_y1", x2="bbox_x2",
        y2="bbox_y2"):
    """Parses an object to generate a bbox payload.
    
    Generates a dict with this structure:
    { "x1": ..., "y1": ..., "x2": ..., "y2": ... }
    @x1, @y1, @x2, @y2 are field names in the object.
    @accessor takes in the object and a field name and gets the value for that
    object.
    """
    return dict_payload_parser(accessor, {
        "x1": x1, "y1": y1, "x2": x2, "y2": y2
    })

def in_array(parser_fn):
    """Generate a new parser function that wraps a payload result in an array.
    """
    return lambda obj: [parser_fn(obj)]

def merge_dict_parsers(parser_fns):
    """Generate a new parser that merges the result of multiple parsers.

    @parser_fns is a list of parser functions. Must parse objects into dicts
    with unique keys.
    """
    # { **obj1, **obj2 } merges the two dicts
    return lambda obj: reduce(lambda obj1, obj2: { **obj1, **obj2 },
            [fn(obj) for fn in parser_fns])

def named_payload(name, parser_fn):
    """Wraps a parser result in a dictionary under given name."""
    return lambda obj: {name: parser_fn(obj)}



"""DEPRECATED. Use rekall.domain_interval_collection instead."""

from rekall.interval_set_3d import Interval3D, IntervalSet3D
from rekall.video_interval_collection import VideoIntervalCollection as VIC
from rekall.interval_set_3d_utils import perf_count
from types import MethodType
import multiprocessing as mp
from tqdm import tqdm

# Functions that ran in worker processes
def _init_workers(context):
    global GLOBAL_CONTEXT
    GLOBAL_CONTEXT = context

def _worker_func_binary(video):
    func, map1, map2 = GLOBAL_CONTEXT
    it1 = map1.get(video, IntervalSet3D([]))
    it2 = map2.get(video, IntervalSet3D([]))
    return (video, func(it1, it2))

def _worker_func_unary(video):
    func, map1 = GLOBAL_CONTEXT
    return (video, func(map1[video]))

class VideoIntervalCollection3D:
    """DEPRECATED. Use DomainIntervalCollection instead.

    A VideoIntervalCollection3D is a wrapper around IntervalSet3D designed for
    videos. Logically, it contains a mapping from video ID's to IntervalSet3D.
    It exposes the same interface as an IntervalList.
    """
    UNARY_METHODS = ["filter_size", "map", "filter", "group_by", "fold_to_set",
            "map_payload", "dilate", "group_by_time", "temporal_coalesce",
            "split"]
    BINARY_METHODS = ["merge", "union", "join", "minus", "filter_against",
            "collect_by_interval"]
    OUT_OF_SYSTEM_UNARY_METHODS = ["size", "empty", "fold", "match"]

    def __new__(cls, *args, **kwargs):
        instance = super(VideoIntervalCollection3D, cls).__new__(cls)
        for method in VideoIntervalCollection3D.UNARY_METHODS:
            setattr(instance, method,
                MethodType(
                VideoIntervalCollection3D._get_wrapped_unary_method(method),
                instance))
        for method in VideoIntervalCollection3D.BINARY_METHODS:
            setattr(instance, method,
                MethodType(
                VideoIntervalCollection3D._get_wrapped_binary_method(method),
                instance))
        for method in VideoIntervalCollection3D.OUT_OF_SYSTEM_UNARY_METHODS:
            setattr(instance, method,
                MethodType(
                VideoIntervalCollection3D\
                        ._get_wrapped_out_of_system_unary_method(method),
                instance))
        return instance

    def __init__(self, video_id_to_intervalset):
        self._video_map = video_id_to_intervalset

    def __repr__(self):
        return "<VideoIntervalCollection3D videos={0}".format(
                self._video_map.keys())

    # Makes this class pickleable
    def __getstate__(self):
        return self._video_map
    def __setstate__(self, video_map):
        self._video_map = video_map


    @classmethod
    def from_iterable(cls, iterable, v_accessor, t_accessor,
            x_accessor=lambda _:(0,1), y_accessor=lambda _:(0,1),
            p_accessor=lambda _:None, progress=False, total=None):
        """
        Construct a VideoIntervalCollection from an iterable collection.

        @iterable is an iterable collection.
        @*_accessor are functions on the item in iterable that returns
            video_id, temporal bounds, spatial bounds, or payload
        """
        video_ids_to_intervals = {}
        for row in (tqdm(iterable, total=total)
                if progress and total is not None else tqdm(iterable)
                if progress else iterable):
            interval = Interval3D(t_accessor(row),
                    x_accessor(row), y_accessor(row), p_accessor(row))
            video_id = v_accessor(row)
            if video_id in video_ids_to_intervals:
                video_ids_to_intervals[video_id].append(interval)
            else:
                video_ids_to_intervals[video_id] = [interval]
        return cls({vid: IntervalSet3D(intervals) for vid, intervals in 
            video_ids_to_intervals.items()})

    @staticmethod
    def django_bbox_default_schema():
        return {
          "video_id": "video_id",
          "t1": "min_frame",
          "t2": "max_frame",
          "x1": "bbox_x1",
          "x2": "bbox_x2",
          "y1": "bbox_y1",
          "y2": "bbox_y2",
          "payload": "id"
        }

    @classmethod
    def from_django_qs(cls, qs, schema={},
            with_payload=None, progress=False, total=None):
        """
        Constructor for a Django queryset.
        By default, the schema is
        {
            "video_id": "video_id",
            "t1": "min_frame",
            "t2": "max_frame",
        }
        """
        schema_final = {
          "video_id": "video_id",
          "t1": "min_frame",
          "t2": "max_frame",
        }
        schema_final.update(schema)
        def get_accessor(field):
            return lambda row: VIC.django_accessor(row, field)
        def get_bounds_accessor(field1, field2):
            return lambda row: (VIC.django_accessor(row, field1),
                                VIC.django_accessor(row, field2))
        s = schema_final
        v = get_accessor(s['video_id'])
        t = get_bounds_accessor(s['t1'], s['t2'])
        kwargs = {}
        if 'x1' in s and 'x2' in s:
            kwargs['x_accessor'] = get_bounds_accessor(s['x1'],s['x2'])
        if 'y1' in s and 'y2' in s:
            kwargs['y_accessor'] = get_bounds_accessor(s['y1'],s['y2'])
        if with_payload is not None:
            kwargs['p_accessor'] = with_payload
        elif 'payload' in s:
            kwargs['p_accessor'] = get_accessor(s['payload'])
        return cls.from_iterable(qs, v, t,
                progress=progress, total=total, **kwargs)

    @staticmethod
    def _remove_empty_intervalsets(video_map):
        new_map = {}
        for video, intervalset in video_map.items():
            if not intervalset.empty():
                new_map[video] = intervalset
        return new_map

    def default_chunksize(self):
        tasks = len(self._video_map)
        return max(1, int(tasks/mp.cpu_count()))


    @staticmethod
    def _get_wrapped_unary_method(name):
        def method(self, *args, parallel=False, profile=False, **kwargs):
            with perf_count(name, profile):
                selfmap = self.get_allintervals()
                videos_to_process = selfmap.keys()
                def func(set1):
                    return getattr(IntervalSet3D, name)(set1,*args,**kwargs)

                if parallel:
                    # Send func selfmap and othermap to the worker processes as
                    # Global variables.
                    with mp.Pool(initializer=_init_workers,
                            initargs=((func,selfmap),)) as pool:
                        videos_results_list = pool.map(
                                _worker_func_unary, videos_to_process,
                                chunksize=self.default_chunksize())
                else:
                    videos_results_list = [
                            (v,func(selfmap[v])) for v in videos_to_process]
            return VideoIntervalCollection3D(
                        {video: results for video, results in 
                            videos_results_list if not results.empty()})
        return method
    
    @staticmethod
    def _get_wrapped_binary_method(name):
        def method(self, other, *args, parallel=False, profile=False, **kwargs):
            with perf_count(name, profile):
                video_map = {}
                selfmap = self.get_allintervals()
                othermap = other.get_allintervals()
                videos_to_process = set(selfmap.keys()).union(othermap.keys())

                def func(set1, set2):
                    return getattr(IntervalSet3D, name)(
                            set1,set2,*args,**kwargs)

                if parallel:
                # Send func selfmap and othermap to the worker processes as
                # Global variables.
                    with mp.Pool(initializer=_init_workers,
                        initargs=((func,selfmap, othermap),)) as pool:
                        videos_results_list = pool.map(
                            _worker_func_binary, videos_to_process,
                            chunksize=self.default_chunksize())
                else:
                    videos_results_list = [
                            (v, func(selfmap.get(v, IntervalSet3D([])),
                                othermap.get(v, IntervalSet3D([])))) for v
                            in videos_to_process]
            return VideoIntervalCollection3D(
                    {video: results for video, results in videos_results_list
                        if not results.empty()})
        return method

    @staticmethod
    def _get_wrapped_out_of_system_unary_method(name):
        def method(self, *args, parallel=False, profile=False, **kwargs):
            with perf_count(name, profile):
                selfmap = self.get_allintervals()
                videos_to_process = selfmap.keys()
                def func(set1):
                    return getattr(IntervalSet3D, name)(set1,*args,**kwargs)

                if parallel:
                    # Send func selfmap and othermap to the worker processes as
                    # Global variables.
                    with mp.Pool(initializer=_init_workers,
                            initargs=((func,selfmap),)) as pool:
                        videos_results_list = pool.map(
                                _worker_func_unary, videos_to_process,
                                chunksize=self.default_chunksize())
                else:
                    videos_results_list = [
                            (v,func(selfmap[v])) for v in videos_to_process]
            return {video: results for video, results in videos_results_list}
        return method

    def get_allintervals(self):
        return self._video_map




"""This class contains various useful merge ops."""

def payload_first(payload1, payload2):
    return payload1

def payload_second(payload1, payload2):
    return payload2

def payload_plus(payload1, payload2):
    return payload1 + payload2

def merge_named_payload(name_to_merge_op):
    """Merging dictionary payload by key.

    name_to_merge_op is a dict mapping from field names to merge_ops.

    Example:
        If name_to_merge_op is 
        {
            'f1': mergeop1,
            'f2': mergeop2,
            'f3': mergeop3
        },
        Then two payloads { 'f1': a1, 'f2': b1, 'f3': c1 } and
        { 'f1': a2, 'f2': b2, 'f3': c2 } will be merged into
        {
          'f1': mergeop1(a1, a2),
          'f2': mergeop2(b1, b2),
          'f3': mergeop3(c1, c2)
        }.
    """

    def merge(p1,p2):
        p = {}
        for name, op in name_to_merge_op.items():
            p[name] = op(p1[name], p2[name])
        return p
    return merge



"""DEPRECATED. Use 3D intervals and IntervalSet3D.match instead.

Spatial predicates on payloads. Assumes that the payload is an array of
objects with at least x1, y1, x2, y2 fields:
[
    {'x1': ..., 'y1': ..., 'x2': ..., 'y2': ..., [other fields, ...]},
    {'x1': ..., 'y1': ..., 'x2': ..., 'y2': ..., [other fields, ...]},
    ...
]

Scene graph Predicates
This is a predicate on the arrangement of objects. See how to create scene
graphs below.
    scene_graph(graph, region, exact) - True if the payload has
        objects in the region that staisfy the scene graph

Constructing scene graphs:
A scene graph defines a predicate on a list of bounding boxes. Scene graphs are
graphs with nodes and edges. Nodes are annotated with unary predicates, and
edges are annotated with binary predicates. A list of bounding boxes satisfies
the scene graph if there is some assignment from bounding boxes to nodes such
that:
    * For each assignment from bounding box to node, the bounding box satisfies
      all the node's unary predicates
    * For each edge, the bounding boxes assigned to the start and end node
      satisfy all the edge's binary predicates

The format of the scene graph is the following:
{
    'nodes': [
        { 'name': 'node1', 'predicates': [pred1, pred2, ...] },
        { 'name': 'node2', 'predicates': [pred1, pred2, ...] },
        ...
    ],
    'edges': [
        { 'start': 'node1', 'end': 'node2', 'predicates': [pred1, pred2, ...] },
        { 'start': 'node1', 'end': 'node3', 'predicates': [pred1, pred2, ...] },
        ...
    ]
}

Constructing regions:
Regions are defined by x1, y1, x2, y2. We have some helper functions for
constructing regions:
    region(x1, y1, x2, y2) - get a region with x1, y1, x2, y2
    full_frame() - get the full frame
    left_half(region=full_frame()) - left half of a region
    right_half(region=full_frame()) - right half of a region
    top_half(region=full_frame()) - top half of a region
    bottom_half(region=full_frame()) - bottom half of a region
    top_left(region=full_frame()) - top left quadrant of a region
    top_right(region=full_frame()) - top right quadrant of a region
    bottom_left(region=full_frame()) - bottom left quadrant of a region
    bottom_right(region=full_frame()) - bottom right quadrant of a region
    center(region=full_frame()) - center of a region (returns a region 1/4 the
        size of the full region, proportional to the full region, and in the
        center of the full region)
"""

from constraint import *

'''Constructing regions'''
def make_region(x1, y1, x2, y2):
    return { 'x1': x1, 'y1': y1, 'x2': x2, 'y2': y2 }
def full_frame():
    return make_region(0., 0., 1., 1.)
def left_half(region=full_frame()):
    return make_region(region['x1'], region['y1'],
            (region['x1'] + region['x2']) / 2., region['y2'])
def right_half(region=full_frame()):
    return make_region((region['x1'] + region['x2']) / 2., region['y1'],
            region['x2'], region['y2'])
def top_half(region=full_frame()):
    return make_region(region['x1'], region['y1'],
            region['x2'], (region['y1'] + region['y2']) / 2.)
def bottom_half(region=full_frame()):
    return make_region(region['x1'], (region['y1'] + region['y2']) / 2.,
            region['x2'], region['y2'])
def top_left(region=full_frame()):
    return left_half(top_half(region))
def top_right(region=full_frame()):
    return right_half(top_half(region))
def bottom_left(region=full_frame()):
    return left_half(bottom_half(region))
def bottom_right(region=full_frame()):
    return right_half(bottom_half(region))
def center(region=full_frame()):
    width = region['x2'] - region['x1']
    height = region['y2'] - region['y1']
    return region(region['x1'] + width / 4.,
            region['y1'] + height / 4.,
            region['x2'] - width / 4.,
            region['y2'] - height / 4.)

'''Scene graph predicates'''
def _region_contains_bbox(region, bbox):
    return (bbox['x1'] >= region['x1'] and
        bbox['x2'] <= region['x2'] and
        bbox['y1'] >= region['y1'] and
        bbox['y2'] <= region['y2'])

def _bboxes_in_region(region, bboxes):
    return [bbox for bbox in bboxes if _region_contains_bbox(region, bbox)]

def _bbox_satisfies_node(bbox, node):
    for pred in node['predicates']:
        if not pred(bbox):
            return False
    return True

def _bboxes_satisfy_edge(bbox1, bbox2, edge):
    for pred in edge['predicates']:
        if not pred(bbox1, bbox2):
            return False
    return True

def scene_graph(graph, region=None, exact=False):
    '''
    Returns a predicate that checks if a list of bounding boxes satisfies the
    scene graph.

    @graph: The scene graph to check against. Has this format:
        {
            'nodes': [
                { 'name': 'node1', 'predicates': [pred1, pred2, ...] },
                { 'name': 'node2', 'predicates': [pred1, pred2, ...] },
                ...
            ],
            'edges': [
                { 'start': 'node1', 'end': 'node2',
                    'predicates': [pred1, pred2, ...] },
                { 'start': 'node1', 'end': 'node3',
                    'predicates': [pred1, pred2, ...] },
                ...
            ]
        }
    @region: Restrict the bounding boxes to this region. If None, then the
        region is the whole screen.
    @exact: If True, the list of bounding boxes must have the same number of
        elements as @graph has nodes.
    '''
    if region is None:
        region = full_frame()

    def compute_predicate(bbox_list):
        bbox_candidates = _bboxes_in_region(region, bbox_list)
        if exact and len(graph['nodes']) is not len(bbox_list):
            return False
        if len(graph['nodes']) == 0:
            return True
        
        '''
        This is a classic constraint satisfaction problem. Our variables are
        the nodes, whose domains are all the bounding boxes that satisfy all
        the node's predicates. Our edges define constraints between variables.
        Satisfying the scene graph is equivalent to satisfying the CSP.
        '''

        problem = Problem()
        variables = []
        for node in graph['nodes']:
            # Pre-compute per-node constraints
            candidates = [i for (i, bbox) in enumerate(bbox_candidates)
                    if _bbox_satisfies_node(bbox, node)]
            if len(candidates) == 0:
                return False
            problem.addVariable(node['name'], candidates)
            variables.append(node['name'])
        problem.addConstraint(AllDifferentConstraint())
        # Add edge constraints.
        # Note that `edge` needs to be explicitly passed into lambda since
        # otherwise it captures `edge` by reference which changes value in
        # the next loop iteration.
        for edge in graph['edges']:
            problem.addConstraint((lambda e: (lambda bbox1, bbox2: _bboxes_satisfy_edge(
                bbox_candidates[bbox1], bbox_candidates[bbox2], e)))(edge),
                (edge['start'], edge['end']))

        return problem.getSolution() is not None

    return compute_predicate




"""DEPRECATED: Use rekall.interval_set_3d instead."""

import itertools
from functools import reduce
from rekall.common import *
from rekall.helpers import *
from rekall.temporal_predicates import *
from rekall.logical_predicates import *
from rekall.merge_ops import *

class Interval:
    """DEPRECATED: Use Interval3D in rekall.interval_set_3d instead.

    A Interval has a start time, end time, and payload.
    """
    def __init__(self, start, end, payload):
        """
        Construct an temporal range from a start time, end time, and integer payload.
        """
        self.start = start
        self.end = end
        self.payload = payload

    def sort_key(intrvl):
        return (intrvl.start, intrvl.end)

    def __repr__(self):
        return "<Interval start:{} end:{} payload:{}>".format(self.start,
                self.end, self.payload)

    def copy(self):
        return Interval(self.start, self.end, self.payload)

    def minus(self, other, payload_merge_op=payload_first):
        """
        Computes the interval difference between self and other and returns results
        in an array.
        If there is no overlap between self and other, [self.copy()] is returned.
        If self is completely contained by other, [] is returned.
        Otherwise, returns a list l of intervals such that the members of l
        maximally cover self without overlapping other.
        The payloads of the members of l are determined by
        payload_merge_op(self, other).
        """
        if overlaps()(self, other):
            payload = payload_merge_op(self.payload, other.payload)
            if during()(self, other) or equal()(self, other):
                return []
            if overlaps_before()(self, other):
                return [Interval(self.start, other.start, payload)]
            if overlaps_after()(self, other):
                return [Interval(other.end, self.end, payload)]
            if during()(other, self):
                return [Interval(self.start, other.start, payload),
                        Interval(other.end, self.end, payload)]
            error_string = "Reached unreachable point in minus with {} and {}"
            error_string = error_string.format(self, other)
            assert False, error_string
        else:
            return [self.copy()]

    def overlap(self, other, payload_merge_op=payload_first):
        """
        Computes the interval overlap between self and other.
        If there is no overlap between self and other, returns None.
        Otherwise, it returns an interval that maximally overlaps both self and
        other, with payload produced by lable_producer_fn(self, other).
        """
        if overlaps()(self, other):
            payload = payload_merge_op(self.payload, other.payload)
            if (during()(self, other) or equal()(self, other)
                    or starts()(self, other) or finishes()(self, other)):
                return Interval(self.start, self.end, payload)
            if overlaps_before()(self, other):
                return Interval(other.start, self.end, payload)
            if overlaps_after()(self, other):
                return Interval(self.start, other.end, payload)
            if (during()(other, self) or starts()(other, self)
                    or finishes()(other, self)):
                return Interval(other.start, other.end, payload)
            error_string = "Reached unreachable point in overlap with {} and {}"
            error_string = error_string.format(self, other)
            assert False, error_string
        else:
            return None

    def merge(self, other, payload_merge_op=payload_first):
        """
        Computes the minimum interval that contains both self and other.
        """
        payload = payload_merge_op(self.payload, other.payload)
        return Interval(min(self.start, other.start),
                max(self.end, other.end), payload)

    def get_start(self):
        return self.start
    def get_end(self):
        return self.end
    def get_payload(self):
        return self.payload
    def length(self):
        return self.end-self.start


class IntervalList:
    """DEPRECATED: Use IntervalSet3D in rekall.interval_set_3d instead.

    A IntervalList is a wrapper around a list of Intervals.

    IntervalList provides core operations map, fold, join, coalesce, and minus.
    See the respective functions for what each does.

    IntervalList also provides sugared functions merge and overlaps, which are
    common ways to use join in a temporal domain.

    OPTIMIZATION: If there are more than 1000 intervals, join operations will
        work over a "working window" equal to 1/100th of the total timespan
        covered by the IntervalList. Instead of computing the full cross
        product for a join and then filtering those pairs by the predicate,
        it only runs the predicate for intervals that differ in time by less
        than the working window. On large IntervalLists, this can increase
        performance dramatically. Users can set the working_window manually
        for all join operations.
    """
    def __init__(self, intrvls):
        if isinstance(intrvls, IntervalList):
            intrvls = intrvls.intrvls
        self.intrvls = sorted([intrvl if isinstance(intrvl, Interval)
                else Interval(intrvl[0], intrvl[1], intrvl[2]) for intrvl in intrvls],
                key = Interval.sort_key)
        if len(self.intrvls) > 0:
            max_end = max([intrvl.end for intrvl in self.intrvls])
            if len(self.intrvls) > 1000:
                self.working_window = (max_end - self.intrvls[0].start) / 100
            else:
                self.working_window = (max_end - self.intrvls[0].start)
        else:
            self.working_window = 0

    def __repr__(self):
        return str(self.intrvls)
    
    # ============== GETTERS ==============
    def size(self):
        """ Gets number of Intervals stored by this IntervalList.
        """
        return len(self.intrvls)

    def get_intervals(self):
        """ Return an ordered list of the Intervals.
        """
        return self.intrvls

    def get_total_time(self):
        """ Get total time. """
        total = 0
        for intrvl in self.intrvls:
            total += intrvl.end - intrvl.start
        return total

    # ============= GENERAL LIST FUNCTIONS =============
    def map(self, map_fn):
        """
        Maps all the intervals in intrvls.

        map_fn takes in an Interval and returns an Interval
        """
        return IntervalList([map_fn(intrvl) for intrvl in self.intrvls])

    def join(self, other, merge_op, predicate, working_window=None):
        """
        Joins self.intrvls with other.intrvls on predicate and produces new
        Intervals based on merge_op.

        merge_op takes in two Intervals and returns a list of Intervals
        predicate takes in two Intervals and returns True or False
        """
        if working_window == None:
            working_window = self.working_window
        output = []
        start_index = 0
        for intrvl in self.intrvls:
            new_start_index = None
            for idx, intrvlother in enumerate(other.intrvls[start_index:]):
                if new_start_index is None:
                    if intrvl.start < intrvlother.start:
                        new_start_index = idx
                    elif intrvl.start - working_window <= intrvlother.end:
                        new_start_index = idx
                if intrvlother.start - working_window > intrvl.end:
                    break
                if predicate(intrvl, intrvlother):
                    new_intrvls = merge_op(intrvl, intrvlother)
                    if len(new_intrvls) > 0:
                        output += new_intrvls
            if new_start_index is not None:
                start_index += new_start_index
        return IntervalList(output)

    def fold(self, fold_fn, init_acc):
        """
        Computes a fold across intrvls.

        fold_fn takes in an accumulator and an Interval.
        """
        return reduce(fold_fn, self.intrvls, init_acc)

    def fold_list(self, fold_fn, init_acc):
        """
        Computes a fold whose accumulator is a list and returns an
        IntervalList.
        """
        return IntervalList(self.fold(fold_fn, init_acc))

    # ============== FUNCTIONS THAT MODIFY SELF ==============
    def coalesce(self, payload_merge_op=payload_first,
            predicate=true_pred(arity=2)):
        """
        Recursively merge all overlapping or touching intervals that satisfy
        predicate. predicate must be an equivalence relation over the intervals
        (must be reflexive, symmetric, and transitive - all the properties of
        a binary relationship that acts like equality in some way).

        """
        if len(self.intrvls) == 0:
            return self
        new_intrvls = []
        current_intervals = []
        for intrvl in self.intrvls:
            for cur in current_intervals:
                # Add any intervals that end before intrvl.start
                if cur.end < intrvl.start:
                    new_intrvls.append(cur)
            current_intervals = [
                cur
                for cur in current_intervals
                if cur.end >= intrvl.start
            ]
            if len(current_intervals) == 0:
                current_intervals.append(intrvl.copy())
                continue
            
            matched_interval = None
            for cur in current_intervals:
                if predicate(cur, intrvl):
                    matched_interval = cur
                    break
            if matched_interval is None:
                current_intervals.append(intrvl)
            else:
                cur.payload = payload_merge_op(cur.payload, intrvl.payload)
                if intrvl.end > cur.end:
                    cur.end = intrvl.end

        for cur in current_intervals:
            new_intrvls.append(cur)

        return IntervalList(new_intrvls)

    def dilate(self, window):
        """
        Expand every temporal range. An temporal range [start, end, i] will turn into
        [start - window, end + window, i].
        """
        return IntervalList(
            [Interval(intrvl.start - window, intrvl.end + window, intrvl.payload) 
                for intrvl in self.intrvls])

    def filter(self, fn):
        """
        Filter every temporal range by fn. fn takes in an Interval and returns true or
        false.
        """
        return IntervalList([intrvl.copy() for intrvl in self.intrvls if fn(intrvl)])

    def filter_length(self, min_length=0, max_length=INFTY):
        """
        Filter temporal ranges so that only temporal ranges of length between min_length and
        max_length are left.
        """
        def filter_fn(intrvl):
            length = intrvl.end - intrvl.start
            return length >= min_length and (max_length == INFTY
                    or length <= max_length)

        return self.filter(filter_fn)

    # ============= FUNCTIONS THAT JOIN WITH ANOTHER INTERVAL LIST =============
    def set_union(self, other):
        """ Combine the temporal ranges in self with the temporal ranges in other.
        """
        return IntervalList(self.intrvls + other.intrvls)

    def filter_against(self, other, predicate=true_pred(arity=2),
            working_window=None):
        """
        Filter the ranges in self against the ranges in other, only keeping the
        ranges in self that satisfy predicate with at least one other range in
        other.
        """
        if working_window == None:
            working_window = self.working_window

        start_index = 0
        output = []
        for intrvl in self.intrvls:
            new_start_index = None
            add_intrvl = False
            for idx, intrvlother in enumerate(other.intrvls[start_index:]):
                if new_start_index is None:
                    if intrvl.start < intrvlother.start:
                        new_start_index = idx
                    elif intrvl.start - working_window <= intrvlother.end:
                        new_start_index = idx
                if intrvlother.start - working_window > intrvl.end:
                    break
                if predicate(intrvl, intrvlother):
                    add_intrvl = True
                    break
            if new_start_index is not None:
                start_index += new_start_index
            if add_intrvl:
                output.append(intrvl.copy())

        return IntervalList(output)

    def minus(self, other, recursive_diff = True, predicate = true_pred(arity=2),
            payload_merge_op = payload_first, working_window=None):
        """
        Calculate the difference between the temporal ranges in self and the temporal ranges
        in other.

        The difference between two intervals can produce up to two new intervals.
        If recursive_diff is True, difference operations will recursively be
        applied to the resulting intervals.
        If recursive_diff is False, the results of each difference operation
        between every valid pair of intervals in self and other will be emitted.

        For example, suppose the following interval is in self:

        |--------------------------------------------------------|

        and that the following two intervals are in other:

                  |--------------|     |----------------|
        
        If recursive_diff is True, this function will produce three intervals:

        |---------|              |-----|                |--------|

        If recursive_diff is False, this function will produce four intervals, some
        of which are overlapping:

        |---------|              |-------------------------------|
        |------------------------------|                |--------|

        Only computes differences for pairs that overlap and that satisfy
        predicate.

        If an interval in self overlaps no pairs in other such that the two
        satisfy predicate, then the interval is reproduced in the output.

        Labels the resulting intervals with payload_merge_op. For recursive_diff,
        the intervals passed in to the payload producer function are the original
        interval and the first interval that touches the output interval.
        """
        if not recursive_diff:
            return self.join(other, lambda intrvl1, intrvl2:
                    intrvl1.minus(intrvl2, payload_merge_op),
                    and_pred(overlaps(), predicate, arity=2),
                    working_window)
        else:
            if working_window == None:
                working_window = self.working_window

            start_index = 0
            output = []
            for intrvl1 in self.intrvls:
                # For each interval in self.intrvls, get all the overlapping
                #   intervals from other.intrvls
                overlapping = []
                new_start_index = None
                for idx, intrvl2 in enumerate(other.intrvls[start_index:]):
                    if new_start_index is None:
                        if intrvl1.start < intrvl2.start:
                            new_start_index = idx
                        elif intrvl1.start - working_window <= intrvl2.end:
                            new_start_index = idx
                    if (overlaps()(intrvl1, intrvl2) and
                        predicate(intrvl1, intrvl2)):
                        overlapping.append(intrvl2)
                if new_start_index is not None:
                    start_index += new_start_index

                if len(overlapping) == 0:
                    output.append(intrvl1.copy())

                # Special case where where intrvl1 has length 0
                if intrvl1.length() == 0 and len(overlapping) is not 0:
                    continue
                
                # Create a sorted list of all start to end points between
                #   intrvl1.start and intrvl1.end, inclusive
                endpoints_set = set([intrvl1.start, intrvl1.end])
                for intrvl in overlapping:
                    if intrvl.start > intrvl1.start:
                        endpoints_set.add(intrvl.start)
                    if intrvl.end < intrvl1.end:
                        endpoints_set.add(intrvl.end)
                endpoints_list = sorted(list(endpoints_set))

                # Calculate longest subsequence endpoint pairs
                longest_subsequences = []
                last_j = -1
                for i in range(len(endpoints_list)):
                    if i <= last_j:
                        continue
                    start = endpoints_list[i]
                    valid = True
                    for intrvl in overlapping:
                        if intrvl.start > start:
                            break
                        if intrvl.start < start and intrvl.end > start:
                            valid = False
                            break
                    if not valid:
                        continue
                    max_j = len(endpoints_list) - 1
                    for j in range(max_j, i, -1):
                        end = endpoints_list[j]
                        intrvl_candidate = Interval(start, end, 0)
                        valid = True
                        for intrvl in overlapping:
                            if intrvl.start > end:
                                break
                            if overlaps()(intrvl, intrvl_candidate):
                                valid = False
                                break
                        if valid:
                            longest_subsequences.append((start, end))
                            last_j = j

                # Figure out which intervals from overlapping to use to
                # construct new intervals
                for subsequence in longest_subsequences:
                    start = subsequence[0]
                    end = subsequence[1]
                    for intrvl in overlapping:
                        if intrvl.end == start or intrvl.start == end:
                            payload = payload_merge_op(intrvl1.payload, intrvl.payload)
                            output.append(Interval(start, end, payload))
                            break

            return IntervalList(output)


    def overlaps(self, other, predicate = true_pred(arity=2), payload_merge_op =
            payload_first, working_window=None):
        """
        Get the overlapping intervals between self and other.

        Only processes pairs that overlap and that satisfy predicate.

        Labels the resulting intervals with payload_merge_op.
        """
        def merge_op(intrvl1, intrvl2):
            return [intrvl1.overlap(intrvl2, payload_merge_op)]
        return self.join(other, merge_op, and_pred(overlaps(), predicate,
            arity=2), working_window)

    def merge(self, other, predicate = true_pred(arity=2), payload_merge_op =
            payload_first, working_window=None):
        """
        Merges pairs of intervals in self and other that satisfy predicate.

        Only processes pairs that satisfy predicate.

        Labels the resulting intervals with payload_merge_op.
        """
        def merge_op(intrvl1, intrvl2):
            return [intrvl1.merge(intrvl2, payload_merge_op)]
        return self.join(other, merge_op, predicate, working_window)




from rekall.temporal_predicates import *
from rekall.merge_ops import *
from rekall.interval_list import Interval, IntervalList
import unittest

class TestInterval(unittest.TestCase):
    def test_minus(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)
        
        intrvl1minusintrvl2 = intrvl1.minus(intrvl2)

        self.assertEqual(len(intrvl1minusintrvl2), 1)
        self.assertEqual(intrvl1minusintrvl2[0].__repr__(),
                "<Interval start:1.0 end:1.5 payload:1>")

        intrvl1minusintrvl2 = intrvl1.minus(intrvl2, payload_merge_op=payload_second)

        self.assertEqual(len(intrvl1minusintrvl2), 1)
        self.assertEqual(intrvl1minusintrvl2[0].__repr__(),
                "<Interval start:1.0 end:1.5 payload:2>")

        intrvl3 = Interval(1.3, 1.6, 5)

        intrvl1minusintrvl3 = intrvl1.minus(intrvl3)
        self.assertEqual(len(intrvl1minusintrvl3), 2)
        self.assertEqual(intrvl1minusintrvl3[0].__repr__(),
                "<Interval start:1.0 end:1.3 payload:1>")
        self.assertEqual(intrvl1minusintrvl3[1].__repr__(),
                "<Interval start:1.6 end:2.0 payload:1>")

        intrvl4 = Interval(0., 5., 1)
        self.assertEqual(intrvl1.minus(intrvl4), [])

    def test_overlap(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)
        
        self.assertEqual(intrvl1.overlap(intrvl2).__repr__(),
                "<Interval start:1.5 end:2.0 payload:1>")
        self.assertEqual(intrvl1.overlap(intrvl2,
            payload_merge_op=payload_second).__repr__(),
                "<Interval start:1.5 end:2.0 payload:2>")

        intrvl3 = Interval(1.3, 1.6, 5)

        self.assertEqual(intrvl1.overlap(intrvl3).__repr__(),
                "<Interval start:1.3 end:1.6 payload:1>")

        intrvl4 = Interval(0., 5., 1)

        self.assertEqual(intrvl1.overlap(intrvl4).__repr__(),
                "<Interval start:1.0 end:2.0 payload:1>")

    def test_merge(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)
        
        self.assertEqual(intrvl1.merge(intrvl2).__repr__(),
                "<Interval start:1.0 end:2.5 payload:1>")
        self.assertEqual(intrvl1.merge(intrvl2,
            payload_merge_op=payload_second).__repr__(),
                "<Interval start:1.0 end:2.5 payload:2>")

        intrvl3 = Interval(5., 7., 1)
        self.assertEqual(intrvl1.merge(intrvl3).__repr__(),
                "<Interval start:1.0 end:7.0 payload:1>")

        intrvl4 = Interval(0., 0.5, 1)
        self.assertEqual(intrvl1.merge(intrvl4).__repr__(),
                "<Interval start:0.0 end:2.0 payload:1>")

class IntervalListTest(unittest.TestCase):
    def test_init(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)

        intrvls1 = IntervalList([intrvl1, intrvl2])
        self.assertEqual(intrvls1.intrvls[0].__repr__(),
                "<Interval start:1.0 end:2.0 payload:1>")
        self.assertEqual(intrvls1.intrvls[1].__repr__(),
                "<Interval start:1.5 end:2.5 payload:2>")

        intrvls1 = IntervalList([intrvl2, intrvl1])
        self.assertEqual(intrvls1.intrvls[0].__repr__(),
                "<Interval start:1.0 end:2.0 payload:1>")
        self.assertEqual(intrvls1.intrvls[1].__repr__(),
                "<Interval start:1.5 end:2.5 payload:2>")

    def test_set_union(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)

        intrvls1 = IntervalList([intrvl1])
        intrvls2 = IntervalList([intrvl2])

        intrvlsu = intrvls1.set_union(intrvls2)
        self.assertEqual(intrvlsu.intrvls[0].__repr__(),
                "<Interval start:1.0 end:2.0 payload:1>")
        self.assertEqual(intrvlsu.intrvls[1].__repr__(),
                "<Interval start:1.5 end:2.5 payload:2>")

        intrvlsu = intrvls2.set_union(intrvls1)
        self.assertEqual(intrvlsu.intrvls[0].__repr__(),
                "<Interval start:1.0 end:2.0 payload:1>")
        self.assertEqual(intrvlsu.intrvls[1].__repr__(),
                "<Interval start:1.5 end:2.5 payload:2>")

    def test_coalesce(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)

        intrvls1 = IntervalList([intrvl1, intrvl2])
        intrvlscoalesced = intrvls1.coalesce()

        self.assertEqual(len(intrvlscoalesced.intrvls), 1)
        self.assertEqual(intrvlscoalesced.intrvls[0].__repr__(),
                "<Interval start:1.0 end:2.5 payload:1>")

        intrvlscoalesced_samepayload = intrvls1.coalesce(
            predicate=lambda i1, i2: i1.payload == i2.payload)
        self.assertEqual(len(intrvlscoalesced_samepayload.intrvls), 2)
        self.assertEqual(intrvlscoalesced_samepayload.intrvls[0].__repr__(),
                "<Interval start:1.0 end:2.0 payload:1>")
        self.assertEqual(intrvlscoalesced_samepayload.intrvls[1].__repr__(),
                "<Interval start:1.5 end:2.5 payload:2>")

    def test_dilate(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)

        intrvls1 = IntervalList([intrvl1, intrvl2]).dilate(0.2)
        self.assertEqual(intrvls1.intrvls[0].__repr__(),
                "<Interval start:0.8 end:2.2 payload:1>")
        self.assertEqual(intrvls1.intrvls[1].__repr__(),
                "<Interval start:1.3 end:2.7 payload:2>")

    def test_filter(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 2.5, 2)

        intrvls1 = IntervalList([intrvl1, intrvl2])

        intrvls1 = intrvls1.filter(lambda intrvl: intrvl.start > 1.1)
        self.assertEqual(len(intrvls1.intrvls), 1)
        self.assertEqual(intrvls1.intrvls[0].__repr__(),
                "<Interval start:1.5 end:2.5 payload:2>")

    def test_filter_length(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(1.5, 3.5, 2)

        intrvls1 = IntervalList([intrvl1, intrvl2])

        intrvls1 = intrvls1.filter_length(min_length=1.1)
        self.assertEqual(len(intrvls1.intrvls), 1)
        self.assertEqual(intrvls1.intrvls[0].__repr__(),
                "<Interval start:1.5 end:3.5 payload:2>")

        intrvls1 = intrvls1.filter_length(max_length=1.8)
        self.assertEqual(len(intrvls1.intrvls), 0)

    def test_filter_against(self):
        intrvl_long1 = Interval(1., 10., 1)
        intrvl_long2 = Interval(3., 15., 2)

        intrvlshort1 = Interval(2., 2.5, 3)
        intrvlshort2 = Interval(2., 2.7, 4)
        intrvlshort3 = Interval(2.9, 3.5, 5)

        intrvlslong = IntervalList([intrvl_long2, intrvl_long1])
        intrvlsshort = IntervalList([
            intrvlshort1,
            intrvlshort2,
            intrvlshort3])

        intrvlsfiltered = intrvlslong.filter_against(intrvlsshort, predicate=during_inv())
        self.assertEqual(len(intrvlsfiltered.intrvls), 1)
        self.assertEqual(intrvlsfiltered.intrvls[0].__repr__(), intrvl_long1.__repr__())

    def test_minus(self):
        intrvl_long1 = Interval(1., 10., 1)
        intrvl_long2 = Interval(3., 15., 2)

        intrvlshort1 = Interval(2., 2.5, 3)
        intrvlshort2 = Interval(2., 2.7, 4)
        intrvlshort3 = Interval(2.9, 3.5, 5)
        intrvlshort4 = Interval(5., 7., 6)
        intrvlshort5 = Interval(9., 12., 7)
        intrvlshort6 = Interval(14., 16., 8)

        intrvlslong = IntervalList([intrvl_long2, intrvl_long1])
        intrvlsshort = IntervalList([
            intrvlshort2,
            intrvlshort5,
            intrvlshort3,
            intrvlshort1,
            intrvlshort4,
            intrvlshort6])

        intrvlsminusrec = intrvlslong.minus(intrvlsshort)
        self.assertEqual(len(intrvlsminusrec.intrvls), 7)
        self.assertEqual(intrvlsminusrec.intrvls[0].__repr__(),
            "<Interval start:1.0 end:2.0 payload:1>")
        self.assertEqual(intrvlsminusrec.intrvls[1].__repr__(),
            "<Interval start:2.7 end:2.9 payload:1>")
        self.assertEqual(intrvlsminusrec.intrvls[2].__repr__(),
            "<Interval start:3.5 end:5.0 payload:1>")
        self.assertEqual(intrvlsminusrec.intrvls[3].__repr__(),
            "<Interval start:3.5 end:5.0 payload:2>")
        self.assertEqual(intrvlsminusrec.intrvls[4].__repr__(),
            "<Interval start:7.0 end:9.0 payload:1>")
        self.assertEqual(intrvlsminusrec.intrvls[5].__repr__(),
            "<Interval start:7.0 end:9.0 payload:2>")
        self.assertEqual(intrvlsminusrec.intrvls[6].__repr__(),
            "<Interval start:12.0 end:14.0 payload:2>")

        intrvlsminusnonrec = intrvlslong.minus(intrvlsshort, recursive_diff = False)
        self.assertEqual(len(intrvlsminusnonrec.intrvls), 15)

        intrvlsminusrec = intrvlslong.minus(intrvlsshort,
                payload_merge_op=payload_second)
        self.assertEqual(len(intrvlsminusrec.intrvls), 7)
        self.assertEqual(intrvlsminusrec.intrvls[0].__repr__(),
            "<Interval start:1.0 end:2.0 payload:3>")
        self.assertEqual(intrvlsminusrec.intrvls[1].__repr__(),
            "<Interval start:2.7 end:2.9 payload:4>")
        self.assertEqual(intrvlsminusrec.intrvls[2].__repr__(),
            "<Interval start:3.5 end:5.0 payload:5>")
        self.assertEqual(intrvlsminusrec.intrvls[3].__repr__(),
            "<Interval start:3.5 end:5.0 payload:5>")
        self.assertEqual(intrvlsminusrec.intrvls[4].__repr__(),
            "<Interval start:7.0 end:9.0 payload:6>")
        self.assertEqual(intrvlsminusrec.intrvls[5].__repr__(),
            "<Interval start:7.0 end:9.0 payload:6>")
        self.assertEqual(intrvlsminusrec.intrvls[6].__repr__(),
            "<Interval start:12.0 end:14.0 payload:7>")

        intrvlsminusrec = intrvlslong.minus(intrvlsshort, predicate=overlaps_before())
        self.assertEqual(len(intrvlsminusrec.intrvls), 2)
        self.assertEqual(intrvlsminusrec.intrvls[0].__repr__(),
            "<Interval start:1.0 end:9.0 payload:1>")
        self.assertEqual(intrvlsminusrec.intrvls[1].__repr__(),
            "<Interval start:3.0 end:14.0 payload:2>")

    def test_minus_against_nothing(self):
        intrvl_long1 = Interval(1., 10., 1)
        intrvl_long2 = Interval(3., 15., 2)
        
        intrvlslong = IntervalList([intrvl_long2, intrvl_long1])

        intrvlshort1 = Interval(20., 20.5, 3)
        intrvlsshort = IntervalList([intrvlshort1])

        intrvlsminusrec = intrvlslong.minus(intrvlsshort)
        self.assertEqual(len(intrvlsminusrec.intrvls), 2)
        self.assertEqual(intrvlsminusrec.intrvls[0].__repr__(),
            "<Interval start:1.0 end:10.0 payload:1>")
        self.assertEqual(intrvlsminusrec.intrvls[1].__repr__(),
            "<Interval start:3.0 end:15.0 payload:2>")


    def test_overlaps(self):
        intrvla1 = Interval(1., 25., 1)
        intrvla2 = Interval(52., 55., 1)
        intrvla3 = Interval(100., 110., 1)
        intrvla4 = Interval(200., 210., 1)
        intrvlb1 = Interval(12., 26., 2)
        intrvlb2 = Interval(50., 53., 2)
        intrvlb3 = Interval(101., 105., 2)
        intrvlb4 = Interval(190., 220., 2)

        intrvlsa = IntervalList([intrvla2, intrvla3, intrvla1, intrvla4])
        intrvlsb = IntervalList([intrvlb2, intrvlb3, intrvlb1, intrvlb4])

        intrvlsoverlap = intrvlsa.overlaps(intrvlsb)
        self.assertEqual(len(intrvlsoverlap.intrvls), 4)
        self.assertEqual(intrvlsoverlap.intrvls[0].__repr__(),
                "<Interval start:12.0 end:25.0 payload:1>")
        self.assertEqual(intrvlsoverlap.intrvls[1].__repr__(),
                "<Interval start:52.0 end:53.0 payload:1>")
        self.assertEqual(intrvlsoverlap.intrvls[2].__repr__(),
                "<Interval start:101.0 end:105.0 payload:1>")
        self.assertEqual(intrvlsoverlap.intrvls[3].__repr__(),
                "<Interval start:200.0 end:210.0 payload:1>")

        intrvlsoverlap = intrvlsb.overlaps(intrvlsa)
        self.assertEqual(len(intrvlsoverlap.intrvls), 4)
        self.assertEqual(intrvlsoverlap.intrvls[0].__repr__(),
                "<Interval start:12.0 end:25.0 payload:2>")
        self.assertEqual(intrvlsoverlap.intrvls[1].__repr__(),
                "<Interval start:52.0 end:53.0 payload:2>")
        self.assertEqual(intrvlsoverlap.intrvls[2].__repr__(),
                "<Interval start:101.0 end:105.0 payload:2>")
        self.assertEqual(intrvlsoverlap.intrvls[3].__repr__(),
                "<Interval start:200.0 end:210.0 payload:2>")

        intrvlsoverlap = intrvlsa.overlaps(intrvlsb, predicate=overlaps_before())
        self.assertEqual(len(intrvlsoverlap.intrvls), 1)
        self.assertEqual(intrvlsoverlap.intrvls[0].__repr__(),
                "<Interval start:12.0 end:25.0 payload:1>")

        intrvlsoverlap = intrvlsa.overlaps(intrvlsb, predicate=overlaps_before(),
                payload_merge_op=payload_second)
        self.assertEqual(len(intrvlsoverlap.intrvls), 1)
        self.assertEqual(intrvlsoverlap.intrvls[0].__repr__(),
                "<Interval start:12.0 end:25.0 payload:2>")

    def test_merge(self):
        intrvla1 = Interval(1., 25., 1)
        intrvla2 = Interval(52., 55., 1)
        intrvla3 = Interval(100., 110., 1)
        intrvla4 = Interval(200., 210., 1)
        intrvlb1 = Interval(12., 26., 2)
        intrvlb2 = Interval(50., 53., 2)
        intrvlb3 = Interval(101., 105., 2)
        intrvlb4 = Interval(190., 220., 2)

        intrvlsa = IntervalList([intrvla2, intrvla3, intrvla1, intrvla4])
        intrvlsb = IntervalList([intrvlb2, intrvlb3, intrvlb1, intrvlb4])

        intrvlsmerge = intrvlsa.merge(intrvlsb, predicate=overlaps())
        self.assertEqual(len(intrvlsmerge.intrvls), 4)
        self.assertEqual(intrvlsmerge.intrvls[0].__repr__(),
                "<Interval start:1.0 end:26.0 payload:1>")
        self.assertEqual(intrvlsmerge.intrvls[1].__repr__(),
                "<Interval start:50.0 end:55.0 payload:1>")
        self.assertEqual(intrvlsmerge.intrvls[2].__repr__(),
                "<Interval start:100.0 end:110.0 payload:1>")
        self.assertEqual(intrvlsmerge.intrvls[3].__repr__(),
                "<Interval start:190.0 end:220.0 payload:1>")

        intrvlsmerge = intrvlsb.merge(intrvlsa, predicate=overlaps())
        self.assertEqual(len(intrvlsmerge.intrvls), 4)
        self.assertEqual(intrvlsmerge.intrvls[0].__repr__(),
                "<Interval start:1.0 end:26.0 payload:2>")
        self.assertEqual(intrvlsmerge.intrvls[1].__repr__(),
                "<Interval start:50.0 end:55.0 payload:2>")
        self.assertEqual(intrvlsmerge.intrvls[2].__repr__(),
                "<Interval start:100.0 end:110.0 payload:2>")
        self.assertEqual(intrvlsmerge.intrvls[3].__repr__(),
                "<Interval start:190.0 end:220.0 payload:2>")

        intrvlsmerge = intrvlsa.merge(intrvlsb, predicate=overlaps_before())
        self.assertEqual(len(intrvlsmerge.intrvls), 1)
        self.assertEqual(intrvlsmerge.intrvls[0].__repr__(),
                "<Interval start:1.0 end:26.0 payload:1>")

        intrvlsmerge = intrvlsa.merge(intrvlsb, predicate=overlaps_before(),
                payload_merge_op=payload_second)
        self.assertEqual(len(intrvlsmerge.intrvls), 1)
        self.assertEqual(intrvlsmerge.intrvls[0].__repr__(),
                "<Interval start:1.0 end:26.0 payload:2>")

        intrvla1 = Interval(1., 25., 1)
        intrvla2 = Interval(52., 55., 1)
        intrvla3 = Interval(100., 110., 1)
        intrvla4 = Interval(200., 210., 1)
        intrvlb1 = Interval(25., 31., 2)
        intrvlb2 = Interval(56., 90., 2)
        intrvlb3 = Interval(101., 105., 2)
        intrvlb4 = Interval(190., 220., 2)

        intrvlsa = IntervalList([intrvla2, intrvla3, intrvla1, intrvla4])
        intrvlsb = IntervalList([intrvlb2, intrvlb3, intrvlb1, intrvlb4])

        intrvlsmerge = intrvlsa.merge(intrvlsb, predicate=meets_before())
        self.assertEqual(len(intrvlsmerge.intrvls), 1)
        self.assertEqual(intrvlsmerge.intrvls[0].__repr__(),
                "<Interval start:1.0 end:31.0 payload:1>")

        intrvlsmerge = intrvlsa.merge(intrvlsb, predicate=before(0.1, 10.0))
        self.assertEqual(len(intrvlsmerge.intrvls), 1)
        self.assertEqual(intrvlsmerge.intrvls[0].__repr__(),
                "<Interval start:52.0 end:90.0 payload:1>")

    def test_map(self):
        intrvla1 = Interval(1., 25., 1)
        intrvla2 = Interval(52., 55., 1)
        intrvla3 = Interval(100., 110., 1)
        intrvla4 = Interval(200., 210., 1)

        intrvlsa = IntervalList([intrvla2, intrvla3, intrvla1, intrvla4])

        intrvlsa = intrvlsa.map(lambda intrvl: Interval(
            intrvl.start + 1, intrvl.end + 1, intrvl.payload))

        self.assertEqual(len(intrvlsa.intrvls), 4)
        self.assertEqual(intrvlsa.intrvls[0].__repr__(),
                "<Interval start:2.0 end:26.0 payload:1>")
        self.assertEqual(intrvlsa.intrvls[1].__repr__(),
                "<Interval start:53.0 end:56.0 payload:1>")
        self.assertEqual(intrvlsa.intrvls[2].__repr__(),
                "<Interval start:101.0 end:111.0 payload:1>")
        self.assertEqual(intrvlsa.intrvls[3].__repr__(),
                "<Interval start:201.0 end:211.0 payload:1>")

    def test_join(self):
        intrvla1 = Interval(1., 25., 1)
        intrvla2 = Interval(52., 55., 1)
        intrvla3 = Interval(100., 110., 1)
        intrvla4 = Interval(200., 210., 1)
        intrvlb1 = Interval(12., 26., 2)
        intrvlb2 = Interval(50., 53., 2)
        intrvlb3 = Interval(101., 105., 2)
        intrvlb4 = Interval(190., 220., 2)

        intrvlsa = IntervalList([intrvla2, intrvla3, intrvla1, intrvla4])
        intrvlsb = IntervalList([intrvlb2, intrvlb3, intrvlb1, intrvlb4])

        def predicate(x, y):
            return x.start == 1. and y.start == 12.

        def merge_op(x, y):
            return [Interval(1., 100., 25)]

        intrvlsudf = intrvlsa.join(intrvlsb, merge_op=merge_op, predicate=predicate)
        self.assertEqual(len(intrvlsudf.intrvls), 1)
        self.assertEqual(intrvlsudf.intrvls[0].__repr__(),
                "<Interval start:1.0 end:100.0 payload:25>")

    def test_fold(self):
        intrvla1 = Interval(1., 25., 1)
        intrvla2 = Interval(52., 55., 1)
        intrvla3 = Interval(100., 110., 1)
        intrvla4 = Interval(200., 210., 1)

        intrvlsa = IntervalList([intrvla2, intrvla3, intrvla1, intrvla4])

        total_payload = intrvlsa.fold(
                lambda acc, intrvl: acc + intrvl.payload, 0.)
        self.assertEqual(total_payload, 4)

        total_length = intrvlsa.fold(
                lambda acc, intrvl: acc + (intrvl.end - intrvl.start), 0.)
        self.assertEqual(total_length, 47.0)

        def fold_fn(acc, intrvl):
            acc.append(intrvl)
            return acc

        intrvlsa = IntervalList(intrvlsa.fold(fold_fn, []))
        self.assertEqual(len(intrvlsa.intrvls), 4)
        self.assertEqual(intrvlsa.intrvls[0].__repr__(),
                "<Interval start:1.0 end:25.0 payload:1>")
        self.assertEqual(intrvlsa.intrvls[1].__repr__(),
                "<Interval start:52.0 end:55.0 payload:1>")
        self.assertEqual(intrvlsa.intrvls[2].__repr__(),
                "<Interval start:100.0 end:110.0 payload:1>")
        self.assertEqual(intrvlsa.intrvls[3].__repr__(),
                "<Interval start:200.0 end:210.0 payload:1>")




"""Managing intervals from different domains.

This module provides DomainIntervalCollection, a dictionary-like class to aid
managing intervals from different domains. We often want to restrict operations
to intervals from the same domain: for example, suppose we use frame number as
the temporal dimension of the interval, it would make little sense to check
temporal overlap between intervals from different videos. In this case,
video_id is the domain of the intervals, and operations on a set of intervals
such as `join` or `minus` should be performed on a per-video basis.

The domain need not be fixed. For example, if the videos are live-broadcasts
and we can use video metadata to recover the absolute UTC time to be the
temporal dimension, it may make sense to re-group the intervals into one
interval set, or use TV channel name as the domain key to re-group them for
downstream processing. DomainIntervalCollection also provides convenient
mechanism for such dynamic re-grouping.

Example Usage:
    collection_A = DomainIntervalCollection.from_django_qs(...)
    collection_B = DomainIntervalCollection(...)
    collection_C = collection_A\\
                   .join(collection_B, ...)\\
                   .map(...)\\
                   .filter(...)\\
                   .union(collection_B.minus(collection_A))
"""
from collections.abc import MutableMapping
from types import MethodType
from tqdm import tqdm

from rekall.interval_set_3d import IntervalSet3D, Interval3D
from rekall.interval_set_3d_utils import perf_count
from rekall.video_interval_collection import VideoIntervalCollection as VIC

def _empty_set():
    return IntervalSet3D([])

class DomainIntervalCollection(MutableMapping):
    """A dictionary from domain key to IntervalSet3D.

    It exposes all the same methods as IntervalSet3D, and delegates the method
    to the underlying IntervalSet3D of each domain in the collection.
    In particular, when calling binary methods such as `join` or `minus` on
    two DomainIntervalCollections, it will perform the method on the underlying
    IntervalSet3Ds separately for each domain.

    For each in-system method of IntervalSet3D (i.e. return value is
    still IntervalSet3D), the corresponding method on DomainIntervalCollection
    returns a DomainIntervalCollection as well.

    For each out-of-system method of IntervalSet3D, namely `size`, `empty`,
    `fold` and `match`, the corresponding method on DomainIntervalCollection
    returns a dictionary from domain key to the result from the underlying
    IntervalSet3D.

    The methods to wrap from IntervalSet3D are defined by the class constants:
    UNARY_METHODS, BINARY_METHODS and OUT_OF_SYSTEM_UNARY_METHODS
    """
    # These methods will return a DomainIntervalCollection where the
    # IntervalSet3D under each group is transformed under the unary operation.
    UNARY_METHODS = ["filter_size", "map", "filter", "group_by", "fold_to_set",
            "map_payload", "dilate", "group_by_time", "temporal_coalesce",
            "split"]
    # These methods takes a second DomainIntervalCollection and 
    # will return a DomainIntervalCollection where the
    # binary operation is performed on the two IntervalSet3Ds with the same key
    BINARY_METHODS = ["merge", "union", "join", "minus", "filter_against",
            "collect_by_interval"]
    # These methods will return a dictionary from domain key to results for 
    # that group.
    OUT_OF_SYSTEM_UNARY_METHODS = ["size", "empty", "fold", "match"]

    def __new__(cls, *args, **kwargs):
        """Creates class instance and adds IntervalSet3D methods on it."""
        instance = super(DomainIntervalCollection, cls).__new__(cls)
        for method in cls.UNARY_METHODS:
            setattr(instance, method,
                MethodType(
                cls._get_wrapped_unary_method(method),
                instance))
        for method in cls.BINARY_METHODS:
            setattr(instance, method,
                MethodType(
                cls._get_wrapped_binary_method(method),
                instance))
        for method in cls.OUT_OF_SYSTEM_UNARY_METHODS:
            setattr(instance, method,
                MethodType(
                cls._get_wrapped_out_of_system_unary_method(method),
                instance))
        return instance

    def __init__(self, grouped_intervals):
        """Initializes with a dictionary from domain key to IntervalSet3D"""
        self._grouped_intervals = grouped_intervals

    def __repr__(self):
        return "<DomainIntervalCollection domains={0}".format(
                self._grouped_intervals.keys())

    # Makes this class pickleable
    def __getstate__(self):
        return self._grouped_intervals
    def __setstate__(self, grouped_intervals):
        self._grouped_intervals = grouped_intervals

    # Dictionary/MutableMapping Interface
    def __getitem__(self, key):
        return self._grouped_intervals.get(key, _empty_set())
    def __setitem__(self, key, value):
        self._grouped_intervals[key] = value
    def __delitem__(self, key):
        del self._grouped_intervals[key]
    def __iter__(self):
        return self._grouped_intervals.__iter__()
    def __len__(self):
        return len(self._grouped_intervals)

    @classmethod
    def from_iterable(cls, iterable, v_accessor, t_accessor,
            x_accessor=lambda _:(0,1), y_accessor=lambda _:(0,1),
            p_accessor=lambda _:None, progress=False, total=None):
        """Constructs a DomainIntervalCollection from an iterable.

        Args:
            iterable: An iterable of arbitrary elements. Each element will
                become an interval in the collection.
            v_accessor: A function that takes an element in iterable and
                returns the domain key for the interval.
            t_accessor: A function that takes an element in iterable and
                returns the temporal bounds for the interval.
            x_accessor (optional): A function that takes an element in iterable
                and returns the x-axis spatial bounds for the interval.
                Defaults to the full relative bounds of [0,1].
            y_accessor (optional): A function that takes an element in iterable
                and returns the y-axis spatial bounds for the interval.
                Defaults to the full relative bounds of [0,1].
            p_accessor (optional): A function that takes an element in iterable
                and returns the payload for the interval.
                Defaults to producing None for all elements.
            progress (Bool, optional): Whether to display a progress bar.
                Defaults to False.
            total (int, optional): Total number of elements in iterable.
                Only used to estimate ETA for the progress bar, and only takes 
                effect if progress is True. 

        Returns:
            A DomainIntervalCollection constructed from iterable and the
            accessors provided.

        Note:
            Everything in iterable will be materialized in RAM.
        """
        key_to_intervals = {}
        for row in (tqdm(iterable, total=total)
                if progress and total is not None else tqdm(iterable)
                if progress else iterable):
            interval = Interval3D(t_accessor(row),
                    x_accessor(row), y_accessor(row), p_accessor(row))
            key = v_accessor(row)
            if key in key_to_intervals:
                key_to_intervals[key].append(interval)
            else:
                key_to_intervals[key] = [interval]
        return cls({key: IntervalSet3D(intervals) for key, intervals in 
            key_to_intervals.items()})

    @classmethod
    def from_django_qs(cls, qs, schema={}, with_payload=None, progress=False):
        """Constructs a DomainIntervalCollection from django QuerySet

        Args:
            qs: A django QuerySet where each record will become an interval
            schema (optional): A dictionary that defines how to transform each
                record to an interval.
                The keys to use are:
                    'domain', 't1', 't2', 'x1', 'x2', 'y1', 'y2', 'payload'
                The value should be a string that is name of the field
                on the django record for the respective attribute of the
                interval.
                If set, it will override per-key the default schema:
                {
                    "domain": "video_id",
                    "t1": "min_frame",
                    "t2": "max_frame"
                }
            with_payload (optional): A function that takes a django record and
                returns the payload for the interval.
                If set, will override the 'payload' field of the schema.
            progress (Bool, optional): whether to display a progress bar.
                Defaults to False.

        Returns:
            A DomainIntervalCollection constructed from django QuerySet.

        Note:
            All records in the QuerySet are materialized in the function call.
        """
        schema_final = {
          "domain": "video_id",
          "t1": "min_frame",
          "t2": "max_frame",
        }
        schema_final.update(schema)
        def get_accessor(field):
            return lambda row: VIC.django_accessor(row, field)
        def get_bounds_accessor(field1, field2):
            return lambda row: (VIC.django_accessor(row, field1),
                                VIC.django_accessor(row, field2))
        s = schema_final
        v = get_accessor(s['domain'])
        t = get_bounds_accessor(s['t1'], s['t2'])
        kwargs = {}
        if 'x1' in s and 'x2' in s:
            kwargs['x_accessor'] = get_bounds_accessor(s['x1'],s['x2'])
        if 'y1' in s and 'y2' in s:
            kwargs['y_accessor'] = get_bounds_accessor(s['y1'],s['y2'])
        if with_payload is not None:
            kwargs['p_accessor'] = with_payload
        elif 'payload' in s:
            kwargs['p_accessor'] = get_accessor(s['payload'])
        total = None
        if progress:
            total = qs.count()
        return cls.from_iterable(qs, v, t,
                progress=progress, total=total, **kwargs)

    @staticmethod
    def django_bbox_default_schema():
        """A default schema for a bounding box record in django database."""
        return {
          "domain": "video_id",
          "t1": "min_frame",
          "t2": "max_frame",
          "x1": "bbox_x1",
          "x2": "bbox_x2",
          "y1": "bbox_y1",
          "y2": "bbox_y2",
          "payload": "id"
        }


    @classmethod
    def from_intervalset(cls, intervalset, key_fn):
        """Constructs a DomainIntervalCollection from an IntervalSet3D.
        
        Args:
            intervalset (IntervalSet3D): An interval set containing all
                intervals to put in the collection.
            key_fn: A function that takes an interval and returns the domain
                key.

        Returns:
            A DomainIntervalCollection with the same intervals organized into
            domains.
        """
        def reducer(acc, interval):
            key = key_fn(interval)
            if key not in acc:
                acc[key] = [interval]
            else:
                acc[key].append(interval)
            return acc
        grouped = intervalset.fold(reducer, {})
        return cls({k:IntervalSet3D(v) for k,v in grouped.items()})

    def get_grouped_intervals(self):
        """Get underlying dictionary from domain key to IntervalSet3D."""
        return self._grouped_intervals

    def get_flattened_intervalset(self):
        """Get an IntervalSet3D containing all intervals in collection."""
        output = []
        for intervals in self.get_grouped_intervals().values():
            output.extend(intervals.get_intervals())
        return IntervalSet3D(output)

    def add_domain_to_payload(self):
        """Adds domain key to payload of each interval in collection

        If each interval in domain K had payload P before, it now has the tuple
            (P, K) as payload.

        Returns:
            A new DomainIntervalCollection with the transformed intervals.

        Note:
            The original DomainIntervalCollection is unchanged. This is the
            same behavior as all unary methods of IntervalSet3D.
        """
        return DomainIntervalCollection({
            k: intervalset.map_payload(lambda p:(p,k))
            for k, intervalset in self.get_grouped_intervals().items()})

    @staticmethod
    def _remove_empty_intervalsets(grouped_intervals):
        new_map = {}
        for key, intervalset in grouped_intervals.items():
            if not intervalset.empty():
                new_map[key] = intervalset
        return new_map

    @staticmethod
    def _get_wrapped_unary_method(name):
        def method(self, *args, profile=False, **kwargs):
            with perf_count(name, profile):
                selfmap = self.get_grouped_intervals()
                keys_to_process = selfmap.keys()

                def func(set1):
                    return getattr(IntervalSet3D, name)(set1,*args,**kwargs)

                results_map = {v:func(selfmap[v]) for v in keys_to_process}
            return DomainIntervalCollection(
                    DomainIntervalCollection._remove_empty_intervalsets(
                        results_map))
        return method

    @staticmethod
    def _get_wrapped_binary_method(name):
        def method(self, other, *args, profile=False, **kwargs):
            with perf_count(name, profile):
                selfmap = self.get_grouped_intervals()
                othermap = other.get_grouped_intervals()
                keys = set(selfmap.keys()).union(othermap.keys())

                def func(set1, set2):
                    return getattr(IntervalSet3D, name)(
                            set1,set2,*args,**kwargs)

                results_map = {v:
                        func(
                            selfmap.get(v, IntervalSet3D([])),
                            othermap.get(v, IntervalSet3D([]))) for v in keys}
            return DomainIntervalCollection(
                    DomainIntervalCollection._remove_empty_intervalsets(
                        results_map))
        return method

    @staticmethod
    def _get_wrapped_out_of_system_unary_method(name):
        def method(self, *args, profile=False, **kwargs):
            with perf_count(name, profile):
                selfmap = self.get_grouped_intervals()
                keys = selfmap.keys()

                def func(set1):
                    return getattr(IntervalSet3D, name)(set1,*args,**kwargs)
            return {v:func(selfmap[v]) for v in keys}
        return method




""" Spatiotemporal 3D interval sets

This module provides:
    Interval3D: Class for a single 3D spatiotemporal volume.
    IntervalSet3D: Class for a set of spatiotemporal volumes.
"""
from rekall.interval_list import Interval
import rekall.interval_set_3d_utils as utils
from rekall.temporal_predicates import *
from rekall.logical_predicates import *
from rekall.merge_ops import * 
from functools import reduce
import constraint as constraint
import copy

class Interval3D:
    """A single 3D spatiotemporal volume

    The volume is defined by bounds on each of the three dimensions and a
    payload of arbitrary type.

    Attributes:
        t: Temporal bounds represented as a pair (start, end).
        x: Spatial bounds along x-axis represented as a pair.
        y: Spatial bounds along y-aixs represented as a pair.
        payload: A metadata field of arbitrary type.
    """
    def __init__(self, t, x=(0,1), y=(0,1), payload=None):
        """Initializes an interval with 3d bounds and payload.

        Args:
            t: Temporal bounds as (start, end) pair.
            x (optional): Spatial bounds along x-axis as a pair.
                Defaults to the full relative bounds from 0 to 1.
            y (optional): Spatial bounds along y-axis as a pair.
                Defaults to the full relative bounds from 0 to 1.
            payload (optional): Metadata of arbitrary type.
                Defaults to None.
        """
        self.t = t
        self.x = x
        self.y = y
        self.payload = payload

    def __repr__(self):
        return ("<Interval3D t: [{0},{1}] "
                "x: [{2},{3}] y: [{4},{5}] payload:{6}>").format(
                self.t[0], self.t[1], self.x[0], self.x[1],
                self.y[0], self.y[1], self.payload)

    def copy(self):
        """Returns a copy of the interval.
        
        Returns:
            An Interval3D that is a copy.

        Notes:
            While the 3D bounds are copied to the return value, the payload
            may only be passed by reference.
        """
        return Interval3D(self.t, self.x,
                self.y, self.payload)

    def combine(self, other, t_combiner, x_combiner, y_combiner,
                payload_merge_op = payload_first):
        """Combines two intervals into one.

        Args:
            other (Interval3D): The other interval to combine with.
            t_combiner: A function that takes two bounds and returns one.
                Used to combine temporal bounds.
            x_combiner: A function that takes two bounds and returns one.
                Used to combine spatial bounds on x-axis.
            y_combiner: A function that takes two bounds and returns one.
                Used to combine spatial bounds on y-axis.
            payload_merge_op: A function that takes the payload in self and
                the payload in other and returns the payload for the combined
                interval.

        Returns:
            A new Interval3D that is the combination of self and other.
        """
        payload = payload_merge_op(self.payload, other.payload)
        t = t_combiner(self.t, other.t)
        x = x_combiner(self.x, other.x)
        y = y_combiner(self.y, other.y)
        return Interval3D(t,x,y,payload)

    def merge(self, other, payload_merge_op=payload_first):
        """Merge two intervals into one by merging bounds in each dimension.

        Merging two bounds means producing a new bound that minimally covers
        both bounds. More formally, the new bound is: 
        (min(start1, start2), max(end1, end2)).
        
        Args:
            other (Interval3D): The other interval to merge with.
            payload_merge_op: A function that takes the payload in self and
                the payload in other and returns the payload for the combined
                interval.

        Returns:
            A new Interval3D that is the merge of self and other.
        """
        return self.combine(other,
                utils.merge_bound,
                utils.merge_bound,
                utils.merge_bound,
                payload_merge_op)

    def overlap_time_merge_space(self, other, payload_merge_op=payload_first):
        """Combines two intervals by overlapping in time and merging in space.
        
        Example:
            {t:(0,2), x:(0,1), y:(1,2)} and {t: (1,3), x:(1,2), y:(0,1)} will
            produce the spatiotemporal volume:
            {t:(1,2), x:(0,2), y:(0,2)}

        Args:
            other (Interval3D): The other interval to combine with.
            payload_merge_op: A function that takes the payload in self and
                the payload in other and returns the payload for the combined
                interval.
        
        Returns:
            A new Interval3D that is the temporal overlap and spatial merge
            of self and other.

        Raises:
            ValueError: If self and other do not overlap in time.
        """
        if utils.T(overlaps())(self, other):
            return self.combine(other,
                    utils.overlap_bound,
                    utils.merge_bound,
                    utils.merge_bound,
                    payload_merge_op)
        else:
            raise ValueError("{0} and {1} do not overlap in time".format(
                self, other))

    def expand_to_frame(self):
        """Expands the interval to the full spatial extent.

        Assuming the spatial bounds are relative bounds within (0,1), this
        operations expands the interval to have the spatial extent of the
        entire frame.

        Returns:
            A new Interval3D with (0,1) as both x- and y-spatial bounds.
        """
        return Interval3D(self.t, (0,1), (0,1), self.payload)

    def length(self):
        """Returns the duration of the interval."""
        return utils.bound_size(self.t)

    def width(self):
        """Returns the length along x-axis of the interval."""
        return utils.bound_size(self.x)

    def height(self):
        """Returns the length along y-axis of the interval."""
        return utils.bound_size(self.y)

    def to_json(self):
        """EXPERIMENTAL. DO NOT USE. Will be removed in the future.

        Returns a python representation of a JSON object
        To customize conversion for payload, add a to_json method
        to the payload.
        """
        return {
            "t": self.t,
            "x": self.x,
            "y": self.y,
            "payload": (self.payload.to_json() if 
                hasattr(self.payload,'to_json') else self.payload)
        }

class IntervalSet3D:
    """A set of Interval3Ds.

    IntervalSet3D provides core operations to transform and combine sets of
    Interval3D such as map, fold, join, union, minus, temporal_coalesce etc.

    Note:
        When the number of intervals is larger than NUM_INTRVLS_THRESHOLD,
        binary operations `join`, `collect_by_interval` and `filter_against`
        will no longer operate on the full cross product of the intervals, and
        instead only look at pairs that are within a temporal neightbourhood
        defined by a time_window that defaults to DEFAULT_FRACTION of the
        overall temporal span of the interval set.
        Each of these operations also has a time_window argument to override
        the default time_window.
    """
    def __init__(self, intrvls):
        """Initializes IntervalSet3D with a list of Interval3Ds.
        
        Args:
            intrvls: a list of Interval3Ds to put in the set.
        """
        self._intrvls = intrvls
        self._sort()
        self._time_window = self._get_time_window()

    def __repr__(self):
        return str(self._intrvls)

    def __len__(self):
        return len(self._intrvls)

    # Sort the intervals in the class.
    def _sort(self):
        self._intrvls.sort(key=utils.sort_key_time_x_y)

    # Compute the default time_window based on the current intervals
    def _get_time_window(self):
        NUM_INTRVLS_THRESHOLD = 1000
        DEFAULT_FRACTION = 1/100
        n = len(self._intrvls)
        if n > 0:
            max_end = max([i.t[1] for i in self._intrvls])
            min_start = self._intrvls[0].t[0]
            if n > NUM_INTRVLS_THRESHOLD:
                return (max_end - min_start) * DEFAULT_FRACTION
            else:
                return max_end - min_start
        else:
            return 0

    def get_intervals(self):
        """Returns a list of Interval3Ds ordered by (t1,t2,x1,x2,y1,y2)."""
        return self._intrvls

    def size(self):
        """Returns the number of intervals in the set."""
        return len(self)

    def empty(self):
        """Returns whether the set is empty."""
        return self.size() == 0

    def map(self, map_fn):
        """Maps a function over all intervals in the set.

        Args:
            map_fn: A function that takes an Interval3D and returns an
                Interval3D

        Returns:
            A new IntervalSet3D with the transformed intervals.
        """
        return IntervalSet3D([map_fn(intrvl) for intrvl in self._intrvls])

    def split(self, split_fn):
        """Splits each interval into a set of intervals.

        Args:
            split_fn: A function that takes an Interval3D and returns an
                IntervalSet3D.

        Returns:
            A new IntervalSet3D with all intervals after the split.
        """
        nested = [split_fn(intrvl).get_intervals() for intrvl in 
                    self._intrvls]
        return IntervalSet3D([i for intrvls in nested for i in intrvls])

    def union(self, other):
        """Set union of two IntervalSet3Ds.

        Args:
            other (IntervalSet3D): the other set to union with.

        Returns:
            A new IntervalSet3D with all intervals in self and other.
        """
        return IntervalSet3D(self._intrvls + other._intrvls)

    def fold(self, reducer, init=None, sort_key=utils.sort_key_time_x_y):
        """Folds a reducer over an ordered list of intervals in the set.
        
        Args:
            reducer: A function that takes a previous state and an interval
                and returns an updated state.
            init (optional): The initial state to use in fold.
                Defaults to None, which means using the first interval as the
                initial state and run reduction from the second interval.
            sort_key (optional): A function that takes an Interval3D and
                returns a value as the sort key that defines the order of the
                list to fold over.
                Default to ascending order by (t1,t2,x1,x2,y1,y2).
        Return:
            The final value of the state after folding all intervals in set.
        """
        lst = self.get_intervals()
        if sort_key != utils.sort_key_time_x_y:
            lst = sorted(lst, key=sort_key)
        if init is None:
            return reduce(reducer, lst)
        else:
            # Avoid taking a reference of the argument
            init = copy.deepcopy(init)
            return reduce(reducer, lst, init)

    def fold_to_set(self, reducer, init=None,
            acc_to_set = lambda acc:IntervalSet3D(acc),
            sort_key = utils.sort_key_time_x_y):
        """Fold over intervals in the set to produce a new IntervalSet3D

        The same as `fold` method except it returns a IntervalSet3D by running
        acc_to_set on the final state of the fold, making it an in-system
        operation.

        Args:
            reducer, init, sort_key: Same as in `fold`.
            acc_to_set (optional): A function that takes the final state of the
                fold and returns a IntervalSet3D.
                Defaults to assuming the state is a list of Interval3Ds and
                constructs a IntervalSet3D on that.
        Returns:
            A new IntervalSet3D that is the result of acc_to_set on the output
            of fold.
        """
        return acc_to_set(self.fold(reducer, init, sort_key))

    def _map_with_other_within_time_window(self, other,
            mapper, time_window=None):
        """Internal helper to deal with pairs of intervals in cross product.

        Map over a list where elements are:
            (interval_in_self, [intervals_in_other])
        where intervals_in_other are those in other that are within time_window
        of interval_in_self. 
        The list is ordered by (t1,t2) of interval_in_self

        Args:
            other (IntervalSet3D): The other IntervalSet3D to do cross product
                with.
            mapper: A function that takes
                (interval_in_self, [intervals_in_other])
                and returns a list of arbitrary type.
            time_window (optional): Restrict interval pairs to those within
                time_window of each other.
                Defaults to None which means using the default time_window
                associated withe self. See class Documentation for more detail.

        Returns:
            A flattend list of mapper outputs.
        """
        if time_window is None:
            time_window = self._time_window
    
        # State is (other_start_index, outputs, done_flag)
        def update_state(state, intrvlself):
            start_index, outputs, done = state
            intervals_in_other = []
            if not done:
                new_start_index = None
                for idx, intrvlother in enumerate(
                        other.get_intervals()[start_index:]):
                    t1, t2 = intrvlself.t
                    v1, v2 = intrvlother.t
                    if t1 - time_window <= v2:
                        if new_start_index is None:
                            new_start_index = idx + start_index
                        if v1 - time_window > t2:
                            break
                        intervals_in_other.append(intrvlother)
                if new_start_index is None:
                    done = True
                else:
                    start_index = new_start_index
            outputs.append(mapper(intrvlself, intervals_in_other))
            return start_index, outputs, done
        state = (0, [], False)
        _,outputs,_ =  self.fold(update_state, state)
        return [r for results in outputs for r in results]

    def join(self, other, predicate, merge_op, time_window=None):
        """Cross-products two sets and combines pairs that pass the predicate.

        Named after the database JOIN operation, this method takes the cross-
        product of two IntervalSet3Ds, filter the resulting set of pairs and 
        form a new IntervalSet3D by combining each pair into a new Interval3D.

        Note:
            `join` does not take the full cross-product but instead only
            consider pairs within time_window of each other. See notes in class
            documentation for more details.

        Args:
            other (IntervalSet3D): The interval set to join with.
            predicate: A function that takes two Interval3Ds and returns a
                bool.
            merge_op: A function that takes two Interval3Ds and returns a list
                of new Interval3Ds.
            time_window (optional): Restrict cross-product to pairs within
                time_window of each other.
                Defaults to None which means using the default time_window
                associated withe self. See class Documentation for more detail.

        Returns:
            A new IntervalSet3D from the intervals produced by merge_op on
            pairs that pass the predicate.
        """
        def map_output(intrvlself, intervals_in_other):
            out = []
            for intrvlother in intervals_in_other:
                if predicate(intrvlself, intrvlother):
                    new_intrvls = merge_op(intrvlself, intrvlother)
                    if len(new_intrvls) > 0:
                        out += new_intrvls
            return out
        return IntervalSet3D(self._map_with_other_within_time_window(
            other, map_output, time_window))

    def filter(self, predicate):
        """Filter the set and keep intervals that pass the predicate.

        Args:
            predicate: A function that takes an Interval3D and returns a bool.

        Returns:
            A new IntervalSet3D which is the filtered set.
        """
        return IntervalSet3D([
            intrvl.copy() for intrvl in self.get_intervals() if 
            predicate(intrvl)])

    def group_by(self, key, merge):
        """Group intervals by key and produces a new interval for each group.

        Args:
            key: A function that takes an Interval3D and returns a value as
                the key to group by.
            merge: A function that takes a group key and an IntervalSet3D of
                the group, and returns one Interval3D.

        Returns:
            A new IntervalSet3D with the results of merge on each group.
        """
        def add_to_group(group, i):
            k = key(i) 
            if k not in group:
                group[k] = [i]
            else:
                group[k].append(i)
            return group
        groups = self.fold(add_to_group, {})
        output = [merge(k, IntervalSet3D(intervals))
            for k, intervals in groups.items()]
        return IntervalSet3D(output)

    def minus(self, other):
        """Set minus on two IntervalSet3Ds.

        Calculate the difference between intervals in self and the
        temporal ranges in other. The spatial dimensions of intervals in self
        are not changed.

        The difference between two intervals can produce up to two new
        intervals.

        For example, suppose the following interval is in self:

        |--------------------------------------------------------|

        and that the following two temporal intervals are in other:

                  |--------------|     |----------------|
        
        this function will produce three intervals:

        |---------|              |-----|                |--------|

        If an interval in self overlaps no pairs in other along the temporal
        dimension, then the interval is reproduced in the output.

        Note:
            Although it is possible to allow set-minus of arbitrary 3D volumes,
            it is unclear how to break the resulting set into bounding volumes.
            Therefore we currently only allow full-frame volumes in `other`,
            essentially only allowing 1D range subtraction on temporal
            dimension.

        Args:
            other (IntervalSet3D): A full-frame IntervalSet3D to subtract from.

        Returns:
            A new IntervalSet3D with the results from set-minus. It may contain
            more intervals than before the subtraction.

        Raises:
            NotImplementedError: If `other` contains intervals without
            full-frame spatial extent.
        """
        # Only allow full-frame intervals in other
        for intrvl in other.get_intervals():
            if intrvl.x != (0,1) or intrvl.y != (0,1):
                raise NotImplementedError
        # Returns a list of intervals that are what is left of intrvl after
        # subtracting all overlapped_intervals.
        def compute_difference(intrvl, overlapped_intervals):
            start = intrvl.t[0]
            overlapped_index = 0
            output = []
            while start < intrvl.t[1]:
                # Each iteration proposes an interval starting at `start`
                # If no overlapped interval goes acoss `start`, then it is
                # a valid start for an interval after the subtraction.
                intervals_across_start = []
                first_interval_after_start = None
                new_overlapped_index = None
                for idx,overlap in enumerate(
                        overlapped_intervals[overlapped_index:]):
                    v1, v2 = overlap.t
                    if new_overlapped_index is None and v2 > start:
                        new_overlapped_index = idx + overlapped_index
                    if v1 <= start and v2 > start:
                        intervals_across_start.append(overlap)
                    elif v1 > start:
                        # overlap is sorted by (t1,t2)
                        first_interval_after_start = overlap
                        break
                if len(intervals_across_start) == 0:
                    # start is valid, now finds an end point
                    if first_interval_after_start is None:
                        end = intrvl.t[1]
                        new_start = end
                    else:
                        end = first_interval_after_start.t[0]
                        new_start = first_interval_after_start.t[1]
                    if end > start:
                        output.append(Interval3D(
                            (start, end), intrvl.x, intrvl.y,
                            intrvl.payload))
                    start = new_start
                else:
                    # start is invalid, now propose another start
                    start = max([i.t[1] for i in intervals_across_start])
                if new_overlapped_index is not None:
                    overlapped_index = new_overlapped_index
            return output
        
        def map_output(intrvl, overlapped):
            # Take only nontrivial overlaps
            to_subtract = [i for i in overlapped if i.length()>0]
            if len(to_subtract) == 0:
                return [intrvl.copy()]
            else:
                return compute_difference(intrvl, to_subtract)

        return IntervalSet3D(self._map_with_other_within_time_window(
            other, map_output, 0))

    def match(self, pattern, exact=False):
        """Pattern matching among the intervals in the set.

        A pattern is a list of constraints, where each constraint is a pair
          (names, predicates)
          A name is a string that names the variable in the constraint.
          A predicate is a function that takes len(names) Interval3Ds as
             arguments and returns True or False.

        A solution is a directionary of assignments, where the key is the name
        of the variable defined in the pattern and the key is the interval in
        the set that gets assigned to that variable. The assignments in a
        solution will satisfy all constraints in the pattern.

        Example:
            pattern = [
                (["harry"], [XY(height_at_least(0.5)), name_is("harry")]),
                (["ron"], [XY(height_at_least(0.5)), name_is("ron")]),
                (["harry","ron"], [XY(same_height()), XY(left_of())]),
            ]

            # solutions are cases in intervalset where harry and ron are at
            # least half the screen height and have the same height and harry
            # is on the left of ron.
            solutions = intervalset.match(pattern)

            # solutions == [
            #   {"harry": Interval3D(...), "ron": Interval3D(...)},
            #   {"harry": Interval3D(...), "ron": Interval3D(...)},
            # ]

        Args:
            pattern: A pattern specified by a list of constraints. See above
                for more details.
            exact (optional): Whether all intervals in the set need to have
                an assignment for it to be considered a solution.
                Defaults to False.

        Returns:
            A list of solutions as described above.
        """
        # Parser for pattern matching
        # pattern: list of (names, constraints)
        # Returns:
        #   A pair of:
        #       Dict of (VariableName, [SingleVariablePredicate])
        #       List of (VariableNames, Predicates)
        def parse_pattern(pattern):
            node_defs = {}
            multi_node_constraints = []
            for names, constraints in pattern:
                if len(names) == 1:
                    name = names[0]
                    if name not in node_defs:
                        node_defs[name] = list(constraints)
                    else:
                        node_defs[name] += constraints
                else:
                    multi_node_constraints.append((names, constraints))
                    for name in names:
                        if name not in node_defs:
                            node_defs[name] = []
            return node_defs, multi_node_constraints

        nodes, constraints = parse_pattern(pattern)
        intervals = self.get_intervals()
        if exact and len(intervals) != len(nodes):
            return []
        if len(nodes) == 0:
            return []

        def satisfies_all(preds, items):
            for pred in preds:
                if not pred(*items):
                    return False
            return True

        def wrap_preds(preds, intervals):
            def pred(*args):
                new_args = [intervals[i] for i in args]
                return satisfies_all(preds, new_args)
            return pred

        prob = constraint.Problem()
        for name, predicates in nodes.items():
            # Pre-compute single variable constraints
            candidates = [i for i, intrvl in enumerate(intervals)
                    if satisfies_all(predicates, [intrvl])]
            if len(candidates) == 0:
                return []
            prob.addVariable(name, candidates)
        prob.addConstraint(constraint.AllDifferentConstraint())
        # Add multi-variable constraints.
        for names, predicates in constraints:
            prob.addConstraint(wrap_preds(predicates, intervals), names)

        solutions = prob.getSolutions()
        if solutions is None:
            return []
        ret = []
        for sol in solutions:
            d = {}
            for key, idx in sol.items():
                d[key] = intervals[idx]
            ret.append(d)
        return ret

    def filter_against(self, other, predicate, time_window=None):
        """Filter intervals in self against intervals in other.

        Keep only intervals in `self` that satisfy the predicate with at least
        one interval in `other`.

        Note:
            We do not check the full cross-product of self and other and
            instead only check pairs that are within time_window of each other.
            See class documentation for more details.

        Args:
            other (IntervalSet3D): The other interval set to filter against.
            predicate: A function that takes one interval in self and an
                interval in other and returns a bool.
            time_window (optional): Restrict cross-product to pairs within
                time_window of each other.
                Defaults to None which means using the default time_window
                associated withe self. See class Documentation for more detail.
        
        Returns:
            A new IntervalSet3D with intervals in self that satisify predicate
            with at least one interval in other.
        """
        def map_output(intrvlself, intrvlothers):
            for intrvlother in intrvlothers:
                if predicate(intrvlself, intrvlother):
                    return [intrvlself.copy()]
            return []
        return IntervalSet3D(self._map_with_other_within_time_window(
            other, map_output, time_window))

    def map_payload(self, fn):
        """Maps a function over payloads of all intervals in the set.

        Args:
            fn: A function that takes the payload of an interval and returns
                the new payload.

        Returns:
            A new IntervalSet3D of intervals with the same 3D bounds but with
            transformed payloads.
        """
        def map_fn(intrvl):
            return Interval3D(intrvl.t,
                    intrvl.x, intrvl.y,
                    fn(intrvl.payload))
        return self.map(map_fn)
    
    @staticmethod
    def _check_axis_value(axis):
        if axis not in {'T','X','Y'}:
            raise ValueError("axis must be one of 'T', 'X', 'Y'")

    def dilate(self, window, axis='T'):
        """Expand the range of every interval in the set along some axis.

        Args:
            window (Number): The amount to extend at each end-point of the
                range. The actual interval will grow by 2*window. Use negative
                number to shrink intervals.
            axis (string, optional): One of 'X', 'Y', 'T'.
                Defaults to 'T'.

        Returns:
            A new IntervalSet3D with the dilated intervals.

        Raises:
            ValueError: If `axis` has invalid value.
        """
        IntervalSet3D._check_axis_value(axis)
        def get_map_fn(window, axis):
            def dilate_bound(b):
                return (b[0]-window, b[1]+window)
            def dilate_t(i):
                return Interval3D(
                        dilate_bound(i.t),
                        i.x, i.y, i.payload)
            def dilate_x(i):
                return Interval3D(i.t, 
                        dilate_bound(i.x),
                        i.y, i.payload)
            def dilate_y(i):
                return Interval3D(i.t, i.x,
                        dilate_bound(i.y),
                        i.payload)
            if axis == 'T':
                return dilate_t
            if axis == 'X':
                return dilate_x
            if axis == 'Y':
                return dilate_y
            raise RuntimeError("Axis not in ['T','X','Y']")
        return self.map(get_map_fn(window, axis))

    def filter_size(self, min_size=0, max_size=INFTY, axis='T'):
        """Filter the intervals by length of the bounds.

        Args:
            min_size (Number, optional): minimum size to pass the filter.
                Defaults to 0.
            max_size (Number, optional): maximum size to pass the filter.
                Defaults to `rekall.common.INFTY`.
            axis (string, optional): One of 'T','X' and 'Y'.
                Defaults to 'T'.

        Returns:
            A new IntervalSet3D of intervals with bound range length within
            [min_size, max_size] along the given axis.

        Raises:
            ValueError: If `axis` has invalid value.
        """
        def getter(axis):
            if axis=='T':
                return Interval3D.length
            if axis=='X':
                return Interval3D.width
            if axis=='Y':
                return Interval3D.height
            raise RuntimeError("Axis not in ['T','X','Y']")
        def get_pred(min_size, max_size, axis):
            func = getter(axis)
            def pred(intrvl):
                val = func(intrvl)
                return val >= min_size and (
                       max_size==INFTY or val<=max_size)
            return pred
        IntervalSet3D._check_axis_value(axis)
        return self.filter(get_pred(min_size, max_size, axis))

    def merge(self, other, predicate, payload_merge_op=payload_first,
            time_window=None):
        """Merges pairs of intervals in self and other that satisfy predicate.

        See documentation on Interval3D.merge for more details on merge.

        Note:
            We do not check the full cross-product of self and other and
            instead only check pairs that are within time_window of each other.
            See class documentation for more details.

        Args:
            other (IntervalSet3D): The other set to merge intervals from.
            predicate: A function that takes one interval in self and one in
                other and returns a bool.
            payload_merge_op (optional): A function that takes a payload from
                self and one from other and returns a new payload.
                Defaults to using payload in self.
            time_window (optional): Restrict cross-product to pairs within
                time_window of each other.
                Defaults to None which means using the default time_window
                associated withe self. See class Documentation for more detail.

        Returns:
            A new IntervalSet3D with the merged intervals.
        """
        def merge_op(i1, i2):
            return [i1.merge(i2, payload_merge_op)]
        return self.join(other, predicate, merge_op, time_window)

    def group_by_time(self):
        """Group intervals by the time span.

        For each group, create a full-frame Interval3D with the same time span,
        and set the group (an IntervalSet3D) to be its payload.

        Returns:
            A new IntervalSet3D of full-frame intervals that have the set of
            intervals in self with the same time span as the payload.
        """
        def key_fn(intrvl):
            return intrvl.t
        def merge_fn(key, intervals):
            return Interval3D(key,
                    payload=intervals)
        return self.group_by(key_fn, merge_fn)

    def collect_by_interval(self, other, predicate,
            filter_empty=True,time_window=None):
        """Collect intervals in other and nest them under intervals in self.

        For each interval in self, nest under it all intervals in other
        that satisfy the predicate.
        The payload of each resulting interval will become a pair
        (T, IntervalSet3D) where T is the original payload in self, and 
        IntervalSet3D is the nested intervals.

        Note:
            We do not check the full cross-product of self and other and
            instead only check pairs that are within time_window of each other.
            See class documentation for more details.

        Args:
            other (IntervalSet3D): Set of intervals to collect.
            predicate: A function that takes one interval in self and one in
                other and returns a bool.
            filter_empty (optional): Whether to remove intervals in self if no
                intervals in other satisfies the predicate with it.
                If False, such intervals will have payload
                (T, IntervalSet3D([])) where T is the original payload.
                Defaults to True.
            time_window (optional): Restrict cross-product to pairs within
                time_window of each other.
                Defaults to None which means using the default time_window
                associated withe self. See class Documentation for more detail.
        
        Returns:
            A new IntervalSet3D of intervals in self but with nested
            interval set from other in the payload.
        """
        def map_output(intrvlself, intrvlothers):
            intrvls_to_nest = IntervalSet3D([
                i for i in intrvlothers if predicate(intrvlself, i)])
            if not intrvls_to_nest.empty() or not filter_empty:
                return [Interval3D(
                    intrvlself.t,
                    intrvlself.x,
                    intrvlself.y,
                    payload=(intrvlself.payload, intrvls_to_nest))]
            return []
        return IntervalSet3D(self._map_with_other_within_time_window(
            other, map_output, time_window))

    def temporal_coalesce(self, payload_merge_op=payload_first, epsilon=0):
        """Recursively merge all temporally touching or overlapping intervals.

        Merge intervals in set if their temporal projections meet, overlap, or
        are up to epsilon apart. Repeat the process until all such intervals
        are merged.

        See documentation on Interval3D.merge for more details on merge.

        Args:
            payload_merge_op (optional): A function that takes two payloads and
                returns a new payload.
                Defaults to selecting the first payload argument.
            epsilon (Number, optional): The slack for judging if intervals
                meet or overlap. Needs to be nonnegative.
                Defaults to 0 (no slack).

        Returns:
            A new IntervalSet3D of intervals that are temporally disjoint and
            are at least epsilon apart.
        """
        def update_output(output, intrvl):
            if len(output)==0:
                output.append(intrvl)
            else:
                merge_candidate = output[-1]
                if utils.T(utils.or_preds(
                    overlaps(),
                    before(max_dist=epsilon)))(merge_candidate, intrvl):
                    output[-1] = merge_candidate.merge(
                            intrvl, payload_merge_op)
                else:
                    output.append(intrvl)
            return output
        return IntervalSet3D(self.fold(update_output, []))

    def to_json(self):
        """EXPERIMENTAL. DO NOT USE. Will be removed in the future.

        Returns a python representation of a JSON object
        """
        return [i.to_json() for i in self._intrvls]



"""DEPRECATED. Use rekall.domain_interval_collection instead."""

from operator import attrgetter
from rekall.interval_list import IntervalList
from rekall.temporal_predicates import *
from rekall.logical_predicates import *
from rekall.helpers import *
from rekall.merge_ops import *
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor
import cloudpickle

class VideoIntervalCollection:
    """DEPRECATED. Use DomainIntervalCollection instead.

    A VideoIntervalCollection is a wrapper around IntervalLists designed for
    videos. Logically, it contains a mapping from video ID's to IntervalLists.
    It exposes the same interface as an IntervalList.
    """

    # ============== CONSTRUCTORS ==============
    def __init__(self, video_ids_to_intervals):
        """
        video_ids_to_intervals is a dict mapping from video ID to either lists
        of (start, end, payload) tuples or an IntervalList.
        """
        self.intervals = {
            video_id: IntervalList(video_ids_to_intervals[video_id])
            for video_id in video_ids_to_intervals
        }

    def from_iterable(iterable, accessor, video_id_field, schema,
            with_payload=None, progress=False, total=None):
        """
        Construct a VideoIntervalCollection from an iterable collection.

        @iterable is an iterable collection.
        @accessor takes items from @iterable and a field name and returns the
        value for that item; i.e. accessor(item, 'id') returns field 'id' of
        item.
        @video_id_field is the field name for the video ID that you group by.
        @schema is a dict mapping from "start", "end", and "payload" to field
        names for those items. For example, @schema might be:
        {
            "start": "min_frame",
            "end": "max_frame",
            "payload": "id"
        }
        @with_payload is for more complex payloads. If with_payload is None,
        then the payload just uses the field name from the schema. Otherwise,
        with_payload is a function that gets a payload from rows in the
        iterable.
        """
        video_ids_to_intervals = {}
        for row in (tqdm(iterable, total=total)
                if progress and total is not None else tqdm(iterable)
                if progress else iterable):
            new_tuple = (
                accessor(row, schema["start"]),
                accessor(row, schema["end"]),
                accessor(row, schema["payload"]) if with_payload is None
                else with_payload(row)
            )
            video_id = accessor(row, video_id_field)
            if accessor(row, video_id_field) in video_ids_to_intervals:
                video_ids_to_intervals[video_id].append(new_tuple)
            else:
                video_ids_to_intervals[video_id] = [new_tuple]

        return VideoIntervalCollection(video_ids_to_intervals)

    def spark_accessor(row, field):
            return row[field]

    def from_spark_df(dataframe, video_id_field="video_id", schema=None,
            with_payload=None, progress=False, total=None):
        """
        Constructor from a Spark dataframe.
        By default, the schema is
        {
            "start": "min_frame",
            "end": "max_frame",
            "payload": "id"
        }
        """
        if schema is None:
            schema = { "start": "min_frame",
                    "end": "max_frame",
                    "payload": "id" }

        dfmaterialized = dataframe.collect()

        return VideoIntervalCollection.from_iterable(dfmaterialized,
                VideoIntervalCollection.spark_accessor, video_id_field, schema,
                with_payload=with_payload, progress=progress, total=total)

    def django_accessor(row, field):
        return attrgetter(field)(row)

    def from_django_qs(qs, video_id_field="video_id", schema=None,
            with_payload=None, progress=False, total=None):
        """
        Constructor for a Django queryset.
        By default, the schema is
        {
            "start": "min_frame",
            "end": "max_frame",
            "payload": "id"
        }
        """
        if schema is None:
            schema = { "start": "min_frame",
                    "end": "max_frame",
                    "payload": "id" }

        return VideoIntervalCollection.from_iterable(qs,
                VideoIntervalCollection.django_accessor, video_id_field,
                schema, with_payload=with_payload, progress=progress, total=total)

    def _remove_empty_intervallists(intervals):
        return { video_id: intervals[video_id]
                for video_id in intervals
                if intervals[video_id].size() > 0 }

    # ============== GETTERS ==============
    def get_intervallist(self, video_id):
        return self.intervals[video_id]
    def get_allintervals(self):
        return self.intervals

    # ============== FUNCTIONS THAT MODIFY SELF ==============
    def coalesce(self, payload_merge_op=payload_first,
            predicate=true_pred(arity=2),
                 parallel=None):
        """ See IntervalList#coalesce for details. """
        if parallel is not None:
            res = self._process_run("coalesce", [payload_merge_op, predicate], parallel)
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in zip(self.intervals.keys(), res)}))
        else:
            return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id: self.intervals[video_id].coalesce(
                            payload_merge_op=payload_merge_op,
                            predicate=predicate)
                        for video_id in list(self.intervals.keys()) }))

    def dilate(self, window, parallel=None):
        """ See IntervalList#dilate for details. """
        if parallel is not None:
            res = self._process_run("dilate", [window], parallel)
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in zip(self.intervals.keys(), res)}))
        else:
            return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id: self.intervals[video_id].dilate(window)
                        for video_id in tqdm(list(self.intervals.keys())) }))

    def filter(self, fn, parallel=None):
        """ See IntervalList#filter for details. """
        if parallel is not None:
            res = self._process_run("filter", [fn], parallel)
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in zip(self.intervals.keys(), res)}))
        else:
            return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id: self.intervals[video_id].filter(fn)
                        for video_id in tqdm(list(self.intervals.keys())) }))

    def _dispatch(arg):
        (intervals, method, args) = arg
        return getattr(intervals, method)(*cloudpickle.loads(args))

    def _process_run(self, method, args, executor):
        args = cloudpickle.dumps(args)
        res = executor.map(
            VideoIntervalCollection._dispatch,
            [(self.intervals[video_id], method, args) for video_id in self.intervals.keys()])
        return list(res)

    def filter_length(self, min_length=0, max_length=INFTY, parallel=None):
        """ See IntervalList#filter_length for details. """
        if parallel is not None:
            res = self._process_run("filter_length", [min_length, max_length], parallel)
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in zip(self.intervals.keys(), res)}))
        else:
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: self.intervals[video_id].filter_length(
                        min_length, max_length)
                    for video_id in list(self.intervals.keys()) }))

    # ============== GENERAL LIST OPERATIONS ==============
    def map(self, map_fn):
        """ See IntervalList#map for details. """
        return VideoIntervalCollection({
            video_id: self.intervals[video_id].map(map_fn)
            for video_id in list(self.intervals.keys()) })

    def join(self, other, merge_op, predicate, working_window=None, parallel=None):
        """
        Inner join on video ID between self and other, and then join the
        IntervalList's of self and other for the video ID.
        """
        if parallel is not None:
            with ProcessPoolExecutor() as executor:
                res = executor.map(
                    VideoIntervalCollection._dispatch,
                    [(self.intervals[video_id], "join", cloudpickle.dumps([other.intervals[video_id], merge_op, predicate, working_window]))
                     for video_id in tqdm(self.intervals.keys())
                     if video_id in list(other.intervals.keys())])
                return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id: r for (video_id, r) in tqdm(zip(self.intervals.keys(), res))
                        if video_id in list(other.intervals.keys())}))
        else:
            return VideoIntervalCollection(
                        VideoIntervalCollection._remove_empty_intervallists({
                            video_id: self.intervals[video_id].join(
                                other.intervals[video_id],
                                merge_op,
                                predicate,
                                working_window)
                            for video_id in list(self.intervals.keys())
                            if video_id in other.intervals }))

    def fold(self, fold_fn, init_acc):
        """
        Returns a dict from video ID to the fold result from applying the
        fold function to each IntervalList. See IntervalList#fold for details.
        """
        return { video_id: self.intervals[video_id].fold(fold_fn, init_acc)
                for video_id in list(self.intervals.keys()) }

    def fold_list(self, fold_fn, init_acc):
        """
        Assumes that the accumulator is a list of Intervals; applies the fold
        function to each IntervalList and returns a Video IntervalCollection.
        """
        return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: self.intervals[video_id].fold_list(
                        fold_fn, init_acc)
                    for video_id in list(self.intervals.keys()) }))

    # ============== FUNCTIONS THAT JOIN WITH ANOTHER COLLECTION ==============

    def set_union(self, other, parallel=None):
        """ Full outer join on video ID's, union between self and other. """
        video_ids = set(self.intervals.keys()).union(
            set(other.intervals.keys()))

        if parallel is not None:
            res = parallel.map(
                VideoIntervalCollection._dispatch,
                [(self.intervals[video_id], "set_union", cloudpickle.dumps([other.intervals[video_id]]))
                 for video_id in tqdm(self.intervals.keys())
                 if video_id in list(other.intervals.keys())])
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in tqdm(zip(self.intervals.keys(), res))
                    if video_id in list(other.intervals.keys())}))


        else:
            return VideoIntervalCollection({
                video_id : (
                    self.intervals[video_id].set_union(other.intervals[video_id])
                    if (video_id in list(self.intervals.keys())
                        and video_id in other.intervals)
                    else (self.intervals[video_id]
                        if video_id in list(self.intervals.keys())
                        else other.intervals[video_id])
                )
                for video_id in video_ids })

    def filter_against(self, other, predicate=true_pred(arity=2),
            working_window=None, parallel=None):
        """
        Inner join on video ID's, computing IntervalList#filter_against.
        """
        if parallel is not None:
            res = parallel.map(
                VideoIntervalCollection._dispatch,
                [(self.intervals[video_id], "filter_against", cloudpickle.dumps([other.intervals[video_id], predicate, working_window]))
                 for video_id in tqdm(self.intervals.keys())
                 if video_id in list(other.intervals.keys())])
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in tqdm(zip(self.intervals.keys(), res))
                    if video_id in list(other.intervals.keys())}))

        else:
            return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id : self.intervals[video_id].filter_against(
                            other.intervals[video_id], predicate, working_window)
                        for video_id in list(self.intervals.keys())
                        if video_id in list(other.intervals.keys()) }))

    def minus(self, other, recursive_diff = True, predicate=true_pred(arity=2),
            payload_merge_op=payload_first, working_window=None, parallel=None):
        """ Left outer join on video ID's, computing IntervalList#minus. """
        if parallel is not None:
            res = parallel.map(
                VideoIntervalCollection._dispatch,
                [(self.intervals[video_id], "minus", cloudpickle.dumps([other.intervals[video_id], recursive_diff, predicate, payload_merge_op, working_window]))
                 for video_id in tqdm(self.intervals.keys())
                 if video_id in list(other.intervals.keys())])
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in tqdm(zip(self.intervals.keys(), res))
                    if video_id in list(other.intervals.keys())}))

        else:
            return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id : (
                            self.intervals[video_id].minus(
                                other.intervals[video_id],
                                recursive_diff = recursive_diff,
                                predicate = predicate,
                                payload_merge_op = payload_merge_op,
                                working_window = working_window)
                            if video_id in other.intervals.keys()
                            else self.intervals[video_id]
                        )
                        for video_id in list(self.intervals.keys()) }))

    def overlaps(self, other, predicate = true_pred(arity=2), payload_merge_op =
            payload_first, working_window=None):
        """ Inner join on video ID's, computing IntervalList#overlaps. """
        return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: self.intervals[video_id].overlaps(
                        other.intervals[video_id], predicate = predicate,
                        payload_merge_op = payload_merge_op,
                        working_window = working_window)
                    for video_id in list(self.intervals.keys())
                    if video_id in list(other.intervals.keys()) }))

    def merge(self, other, predicate = true_pred(arity=2), payload_merge_op =
            payload_first, working_window=None, parallel=None):
        """ Inner join on video ID's, computing IntervalList#merge. """
        if parallel is not None:
            res = parallel.map(
                VideoIntervalCollection._dispatch,
                [(self.intervals[video_id], "merge", cloudpickle.dumps([other.intervals[video_id], predicate, payload_merge_op, working_window]))
                 for video_id in tqdm(self.intervals.keys())
                 if video_id in list(other.intervals.keys())])
            return VideoIntervalCollection(
                VideoIntervalCollection._remove_empty_intervallists({
                    video_id: r for (video_id, r) in tqdm(zip(self.intervals.keys(), res))
                    if video_id in list(other.intervals.keys())}))
        else:
            return VideoIntervalCollection(
                    VideoIntervalCollection._remove_empty_intervallists({
                        video_id: self.intervals[video_id].merge(
                            other.intervals[video_id], predicate = predicate,
                            payload_merge_op = payload_merge_op,
                            working_window = working_window)
                        for video_id in list(self.intervals.keys())
                        if video_id in list(other.intervals.keys()) }))



from rekall.payload_predicates import *
from rekall.bbox_predicates import *
from rekall.spatial_predicates import *
from rekall.interval_list import Interval
import unittest

class TestSpatialPredicates(unittest.TestCase):
    def setUp(self):
        self.bboxes = [
            { 'x1': 0.1, 'y1': 0.1, 'x2': 0.3, 'y2': 0.5 },    
            { 'x1': 0.4, 'y1': 0.1, 'x2': 0.6, 'y2': 0.5 },    
            { 'x1': 0.7, 'y1': 0.1, 'x2': 0.9, 'y2': 0.5 }    
        ]

    def test_scene_node_constraints(self):
        one_box_graph = {
            'nodes': [
                { 'name': 'box1', 'predicates': [] }
            ],
            'edges': []
        }
        self.assertTrue(scene_graph(one_box_graph)(self.bboxes))
        self.assertFalse(scene_graph(one_box_graph, exact=True)(self.bboxes))
        self.assertFalse(scene_graph(
            one_box_graph, 
            region=left_half(), 
            exact=True)(self.bboxes))

        one_box_graph['nodes'][0]['predicates'] = [
            position(0.4, 0.1, 0.6, 0.5)
        ]
        self.assertTrue(scene_graph(one_box_graph)(self.bboxes))

        one_box_graph['nodes'][0]['predicates'] = [
            area_at_least(0.1)
        ]
        self.assertFalse(scene_graph(one_box_graph)(self.bboxes))

    def test_scene_graph_edge_constraints(self):
        three_boxes_graph = {
            'nodes': [
                { 'name': 'box1', 'predicates': [] },
                { 'name': 'box2', 'predicates': [] },
                { 'name': 'box3', 'predicates': [] }
            ],
            'edges': []
        }
        self.assertTrue(scene_graph(three_boxes_graph)(self.bboxes))
        self.assertTrue(
                scene_graph(three_boxes_graph, exact=True)(self.bboxes))
        self.assertTrue(scene_graph(
            three_boxes_graph, 
            region=top_half())(self.bboxes))
        self.assertTrue(scene_graph(
            three_boxes_graph, 
            region=top_half(),
            exact=True)(self.bboxes))
        
        self.assertFalse(scene_graph(
            three_boxes_graph, 
            region=bottom_half())(self.bboxes))
        self.assertFalse(scene_graph(
            three_boxes_graph, 
            region=top_left())(self.bboxes))

        three_boxes_graph['edges'] = [
            { 'start': 'box1', 'end': 'box2', 'predicates': [same_area()] },
            { 'start': 'box1', 'end': 'box3', 'predicates': [same_area()] },
            { 'start': 'box2', 'end': 'box3', 'predicates': [same_area()] }
        ]
        self.assertTrue(scene_graph(three_boxes_graph)(self.bboxes))
        self.assertTrue(scene_graph(three_boxes_graph,
            exact=True)(self.bboxes))
        self.assertTrue(scene_graph(three_boxes_graph,
            region=top_half())(self.bboxes))
        self.assertTrue(scene_graph(three_boxes_graph,
            region=top_half(), exact=True)(self.bboxes))
        
        self.assertFalse(scene_graph(three_boxes_graph,
            region=bottom_half())(self.bboxes))
        self.assertFalse(scene_graph(three_boxes_graph,
            region=top_left())(self.bboxes))

        three_boxes_graph['edges'] = [
            { 'start': 'box1', 'end': 'box2', 'predicates': [left_of()] },
            { 'start': 'box1', 'end': 'box3', 'predicates': [left_of()] },
            { 'start': 'box2', 'end': 'box3', 'predicates': [left_of()] }
        ]
        self.assertTrue(scene_graph(three_boxes_graph)(self.bboxes))

        three_boxes_graph['edges'] = [
            { 'start': 'box1', 'end': 'box2', 'predicates': [left_of()] },
            { 'start': 'box1', 'end': 'box3', 'predicates': [left_of()] },
            { 'start': 'box2', 'end': 'box3', 'predicates': [left_of(), right_of()] }
        ]
        self.assertFalse(scene_graph(three_boxes_graph)(self.bboxes))

    def test_scene_node_and_edge_constraints(self):
        three_boxes_graph = {
            'nodes': [
                { 'name': 'box1', 'predicates': [height_at_least(0.3)] },
                { 'name': 'box2', 'predicates': [height_at_least(0.3)] },
                { 'name': 'box3', 'predicates': [height_at_least(0.3)] }
            ],
            'edges': [
                { 'start': 'box1', 'end': 'box2', 'predicates': [same_area()] },
                { 'start': 'box1', 'end': 'box3', 'predicates': [same_area()] },
                { 'start': 'box2', 'end': 'box3', 'predicates': [same_area()] }
            ]
        }
        self.assertTrue(scene_graph(three_boxes_graph)(self.bboxes))
        self.assertTrue(
                scene_graph(three_boxes_graph, exact=True)(self.bboxes))
        self.assertTrue(scene_graph(
            three_boxes_graph, 
            region=top_half())(self.bboxes))
        self.assertTrue(scene_graph(
            three_boxes_graph, 
            region=top_half(),
            exact=True)(self.bboxes))

        three_boxes_graph['nodes'] = [
            { 'name': 'box1', 'predicates': [has_value('x1', 0.1)] },
            { 'name': 'box2', 'predicates': [has_value('x1', 0.4)] },
            { 'name': 'box3', 'predicates': [has_value('x1', 0.7)] }
        ]
        self.assertTrue(scene_graph(three_boxes_graph)(self.bboxes))
        self.assertTrue(
                scene_graph(three_boxes_graph, exact=True)(self.bboxes))
        self.assertTrue(scene_graph(
            three_boxes_graph, 
            region=top_half())(self.bboxes))
        self.assertTrue(scene_graph(
            three_boxes_graph, 
            region=top_half(),
            exact=True)(self.bboxes))

        three_boxes_graph['edges'] = [
            { 'start': 'box1', 'end': 'box2', 'predicates': [left_of()] },
            { 'start': 'box1', 'end': 'box3', 'predicates': [left_of()] },
            { 'start': 'box2', 'end': 'box3', 'predicates': [right_of()] }
        ]
        self.assertFalse(scene_graph(three_boxes_graph)(self.bboxes))



from rekall.payload_predicates import *
from rekall.list_predicates import *
from rekall.interval_list import Interval
import unittest

class TestListPredicates(unittest.TestCase):
    def test_cardinal_predicates(self):
        bboxes = [
            { 'x1': 0.1, 'y1': 0.1, 'x2': 0.3, 'y2': 0.5 },    
            { 'x1': 0.4, 'y1': 0.1, 'x2': 0.6, 'y2': 0.5 },    
            { 'x1': 0.7, 'y1': 0.1, 'x2': 0.9, 'y2': 0.5 }    
        ]
        
        self.assertTrue(length_exactly(3)(bboxes))
        self.assertTrue(length_at_least(3)(bboxes))
        self.assertTrue(length_at_most(3)(bboxes))
        self.assertTrue(length_between(3, 5)(bboxes))
        self.assertFalse(length_exactly(2)(bboxes))
        self.assertFalse(length_at_least(4)(bboxes))
        self.assertFalse(length_at_most(2)(bboxes))
        self.assertFalse(length_between(1, 2)(bboxes))




""" Face landmark predicates.

Each of these predicates takes in an object that is assumed to have the 
following structure:
{
    'landmarks': obj
}
where obj has the following functions, each of which returns an (N, 2) numpy
array containing all the landmark points:
    face_outline()
    right_eyebrow()
    left_eyebrow()
    nose_bridge()
    nose_bottom()
    right_eye()
    left_eye()
    outer_lips()
    inner_lips()

Simple predicates we have so far:
    looking_left - calculated based on nose direction
    looking_right - calculated based on nose direction
"""

import numpy as np

def looking_left(landmarks):
    """Whether the face is looking to the left.

    Checks whether the bottom landmark on the nose is left of the middle
    of the nose bottom.
    """
    nose_bridge = landmarks['landmarks'].nose_bridge()
    nose_bottom = landmarks['landmarks'].nose_bottom()

    nose_bottom_x_min = np.min(nose_bottom[:,0])
    nose_bottom_x_max = np.max(nose_bottom[:,0])

    return nose_bridge[-1][0] < (nose_bottom_x_min + nose_bottom_x_max) / 2

def looking_right(landmarks):
    """Whether the face is looking to the right.

    Checks whether the bottom landmark on the nose is right of the middle
    of the nose bottom.
    """
    nose_bridge = landmarks['landmarks'].nose_bridge()
    nose_bottom = landmarks['landmarks'].nose_bottom()

    nose_bottom_x_min = np.min(nose_bottom[:,0])
    nose_bottom_x_max = np.max(nose_bottom[:,0])

    return nose_bridge[-1][0] > (nose_bottom_x_min + nose_bottom_x_max) / 2



from rekall.temporal_predicates import *
from rekall.interval_list import Interval
import unittest

class TestTemporalPredicates(unittest.TestCase):
    def test_before(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(3., 4., 1)
        self.assertTrue(before()(intrvl1, intrvl2))
        self.assertFalse(before()(intrvl2, intrvl1))

        intrvl3 = Interval(2., 4., 1)
        self.assertTrue(before()(intrvl1, intrvl3))

    def test_before_range(self):
        pred = before(10., 20.)
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(3., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(12., 15., 1)
        intrvl4 = Interval(22.5, 25., 1)
        self.assertTrue(pred(intrvl1, intrvl3))
        self.assertFalse(pred(intrvl1, intrvl4))
        self.assertTrue(pred(intrvl2, intrvl4))

    def test_after(self):
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(3., 4., 1)
        self.assertFalse(after()(intrvl1, intrvl2))
        self.assertTrue(after()(intrvl2, intrvl1))

        intrvl3 = Interval(2., 4., 1)
        self.assertTrue(after()(intrvl3, intrvl1))
            
    def test_after_range(self):
        pred = after(10., 20.)
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(3., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(12., 15., 1)
        intrvl4 = Interval(22.5, 25., 1)
        self.assertTrue(pred(intrvl3, intrvl1))
        self.assertFalse(pred(intrvl4, intrvl1))
        self.assertTrue(pred(intrvl4, intrvl2))

    def test_overlaps(self):
        pred = overlaps()
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(3., 4., 1)

        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))

        intrvl3 = Interval(3.5, 4.5, 1)
        intrvl4 = Interval(2.5, 3.5, 1)
        self.assertTrue(pred(intrvl2, intrvl3))
        self.assertTrue(pred(intrvl2, intrvl4))

        intrvl5 = Interval(2.5, 4.5, 1)
        intrvl6 = Interval(3.3, 3.6, 1)
        self.assertTrue(pred(intrvl2, intrvl5))
        self.assertTrue(pred(intrvl2, intrvl6))

        intrvl7 = Interval(3., 3.5, 1)
        intrvl8 = Interval(3., 4.5, 1)
        self.assertTrue(pred(intrvl2, intrvl7))
        self.assertTrue(pred(intrvl2, intrvl8))

        intrvl9 = Interval(3.3, 4., 1)
        intrvl10 = Interval(2.1, 4., 1)
        self.assertTrue(pred(intrvl2, intrvl9))
        self.assertTrue(pred(intrvl2, intrvl10))

        intrvl11 = Interval(2.3, 3., 1)
        intrvl12 = Interval(4., 4.5, 1)
        self.assertFalse(pred(intrvl2, intrvl11))
        self.assertFalse(pred(intrvl2, intrvl12))

        self.assertTrue(pred(intrvl2, intrvl2))

    def test_overlapsbefore(self):
        pred = overlaps_before()
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertTrue(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(2., 5., 1)
        intrvl4 = Interval(2., 3., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_overlapsafter(self):
        pred = overlaps_after()
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertTrue(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_starts(self):
        pred = starts()
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(2., 5., 1)
        intrvl4 = Interval(2., 3., 1)
        self.assertTrue(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_startsinv(self):
        pred = starts_inv()
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(2., 5., 1)
        intrvl4 = Interval(2., 3., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertTrue(pred(intrvl2, intrvl4))

    def test_finishes(self):
        pred = finishes()
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertTrue(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_finishesinv(self):
        pred = finishes_inv()
        intrvl1 = Interval(1., 3., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))
        
        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertTrue(pred(intrvl2, intrvl4))

    def test_during(self):
        pred = during()
        intrvl1 = Interval(3., 3.5, 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertTrue(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))

        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_duringinv(self):
        pred = during_inv()
        intrvl1 = Interval(3., 3.5, 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertTrue(pred(intrvl2, intrvl1))

        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_meetsbefore(self):
        pred = meets_before()
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertTrue(pred(intrvl1, intrvl2))
        self.assertFalse(pred(intrvl2, intrvl1))

        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_meetsafter(self):
        pred = meets_after()
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(2., 4., 1)
        self.assertFalse(pred(intrvl1, intrvl2))
        self.assertTrue(pred(intrvl2, intrvl1))

        intrvl3 = Interval(1., 4., 1)
        intrvl4 = Interval(2.5, 4., 1)
        self.assertFalse(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl2, intrvl4))

    def test_equal(self):
        pred = equal()
        intrvl1 = Interval(1., 2., 1)
        intrvl2 = Interval(2., 4., 1)
        intrvl3 = Interval(2., 4., 5)
        self.assertTrue(pred(intrvl1, intrvl1))
        self.assertTrue(pred(intrvl2, intrvl2))
        self.assertTrue(pred(intrvl3, intrvl3))
        self.assertTrue(pred(intrvl2, intrvl3))
        self.assertFalse(pred(intrvl1, intrvl2))




"""Predicates on lists.

Right now just predicates based on length.
    length_exactly(n) - True if length = n
    length_at_least(n) - True if length >= n
    length_at_most(n) - True if length <= n
    length_between(n1, n2) - True if n1 <= length <= n2
"""

def length_exactly(n):
    return lambda l: len(l) == n
def length_at_least(n):
    return lambda l: len(l) >= n
def length_at_most(n):
    return lambda l: len(l) <= n
def length_between(n1, n2):
    return lambda l: (len(l) >= n1 and len(l) <= n2)



"""Utilities for working with 3D Intervals.

Bounds-related:
    merge_bound: Merges two 1D ranges.
    overlap_bound: Overlaps two 1D ranges.
    bound_size: Length of the 1D range.

Default sortkey:
    sort_key_time_x_y: Returns tuple (t1,t2,x1,x2,y1,y2) from 3D Interval

Predicate-related:
    T: Adapter for 1D-range predicates on temporal dimension of Interval3D(s).
    X: Adapter for 1D-range predicates on x-dimension of Interval3D(s).
    Y: Adapter for 1D-range predicates on y-dimension of Interval3D(s).
    XY: Adapter for bounding-box predicates on xy-projection of Interval3D(s).
    P: Adapter for predicates on payload of Interval3D(s).
    not_pred: Transforms the predicate to output its negation.
    and_preds: Composes predciates by ANDing them.
    or_preds: Composes predicates by ORing them.

Performance-related:
    perf_count: ContextManager that prints the walltime to STDOUT.
"""

from rekall.interval_list import Interval
from contextlib import contextmanager
from time import perf_counter

# Bound Combiners
def merge_bound(bound1, bound2):
    """Merges two 1D bounds.
    
    Args:
        bound1, bound2: Bounds to merge. They are in the form of pair
        (start, end).
    
    Returns:
        The merged bound.
    """
    return (min(bound1[0], bound2[0]),max(bound1[1], bound2[1]))

def overlap_bound(bound1, bound2):
    """Overlaps two 1D bounds.
    
    Args:
        bound1, bound2: Bounds to overlap. They are in the form of pair
        (start, end). They need to overlap for the return value to be a valid
        bound.
    
    Returns:
        The overlap between two bounds.
    """
    return (max(bound1[0], bound2[0]),min(bound1[1], bound2[1]))

def bound_size(b):
    """Length of the given bound."""
    return b[1]-b[0]

# Sortkeys for Interval3D
def sort_key_time_x_y(interval):
    """A sortkey function that sorts by (t1,t2,x1,x2,y1,y2).

    (t1,t2), (x1,x2), (y1,y2) are the 3D bounds on the interval.

    Args:
        interval (Interval3D): the interval to compute sort key for.

    Returns:
        6-Tuple that is the sortkey.
    """
    return (*interval.t,
            *interval.x,
            *interval.y)

# Adapters for spatiotemporal predicates
def T(pred):
    """Adapts predicate to work on temporal range of Interval3D(s).

    Example:
        from rekall.temporal_predicates import overlaps
        # Whether two intervals overlap temporally.
        T(overlaps())(interval1, interval2)

    Args:
        pred: A predicate on some number of 1d ranges defined by
            `start` and `end` attributes on them.

    Returns:
        A predicate on temporal ranges of same number of Interval3Ds
    """
    def new_pred(*interval_3ds):
        return pred(*[
            Interval(i.t[0], i.t[1], i.payload) for i in interval_3ds
        ])
    return new_pred

def X(pred):
    """Adapts predicate to work on spatial x-dimension of Interval3D(s).

    Example:
        from rekall.temporal_predicates import overlaps
        # Whether two intervals overlap horizontally.
        X(overlaps())(interval1, interval2)

    Args:
        pred: A predicate on some number of 1d ranges defined by
            `start` and `end` attributes on them.

    Returns:
        A predicate on x-ranges of same number of Interval3Ds
    """
    def new_pred(*interval_3ds):
        return pred(*[
            Interval(i.x[0], i.x[1], i.payload) for i in interval_3ds
        ])
    return new_pred

def Y(pred):
    """Adapts predicate to work on spatial y-dimension of Interval3D(s).

    Example:
        from rekall.temporal_predicates import overlaps
        # Whether two intervals overlap vertically.
        Y(overlaps())(interval1, interval2)

    Args:
        pred: A predicate on some number of 1d ranges defined by
            `start` and `end` attributes on them.

    Returns:
        A predicate on y-ranges of same number of Interval3Ds
    """
    def new_pred(*interval_3ds):
        return pred(*[
            Interval(i.y[0], i.y[1], i.payload) for i in interval_3ds
        ])
    return new_pred

def _interval_3d_to_bbox(intrvl):
    return {'x1':intrvl.x[0], 'x2':intrvl.x[1],
            'y1':intrvl.y[0], 'y2':intrvl.y[1],
            'payload': intrvl.payload
            }

def XY(pred):
    """Adapts predicate to work on spatial dimensions of Interval3D(s).

    Example:
        from rekall.bbox_predicates import left_of
        # Whether interval1 is to the left of interval2 
        XY(left_of())(interval1, interval2)

    Args:
        pred: A predicate on some number of bounding boxes, each represented
            as a dictionary with keys 'x1','x2','y1','y2'.

    Returns:
        A predicate on xy-projections of same number of Interval3Ds
    """
    def new_pred(*interval_3ds):
        return pred(*[
            _interval_3d_to_bbox(i) for i in interval_3ds 
        ])
    return new_pred

def P(pred):
    """Adapts predicate to work on payloads of Interval3D(s).

    Example:
        # frames_with_faces are set of frames where each frame has a payload
        # that is an IntervalSet3D of faces in that frame.
        # To get frames whose faces satisfy pattern 
        frames_with_faces.filter(P(lambda faces: len(faces.match(pattern))>0))

    Args:
        pred: A predicate on some number of payloads

    Returns:
        A predicate on payloads of same number of Interval3Ds
    """
    def new_pred(*interval_3ds):
        return pred(*[
            i.payload for i in interval_3ds 
        ])
    return new_pred


# Adapters for logical combinations of predicates
def not_pred(pred):
    """Negates the predicate"""
    def new_pred(*args):
        return not pred(*args)
    return new_pred

def and_preds(*preds):
    """ANDs the predicates"""
    def new_pred(*args):
        for pred in preds:
            if not pred(*args):
                return False
        return True
    return new_pred

def or_preds(*preds):
    """ORs the predicates"""
    def new_pred(*args):
        for pred in preds:
            if pred(*args):
                return True
        return False
    return new_pred

# Performance profling util
@contextmanager
def perf_count(name, enable=True):
    """Prints wall time for the code block to STDOUT

    Example:
        with perf_count("test code"):
            sleep(10)
        # Writes to stdout:
        # test code starts.
        # test code ends after 10.01 seconds
    """
    if not enable:
        yield
    else:
        print("{0} starts.".format(name))
        s = perf_counter()
        yield
        t = perf_counter()
        print("{0} ends after {1:.2f} seconds".format(name, t-s))



import unittest

from rekall.interval_set_3d import Interval3D, IntervalSet3D
from rekall.runtime import (Runtime, get_forked_process_pool_factory,
        get_spawned_process_pool_factory, RekallRuntimeException)
from rekall.domain_interval_collection import DomainIntervalCollection

class TestRuntime(unittest.TestCase):
    @staticmethod
    def dummy_intervalset():
        return IntervalSet3D([TestRuntime.dummy_interval()])

    @staticmethod
    def query(vids):
        return DomainIntervalCollection({
            vid: TestRuntime.dummy_intervalset() for vid in vids })

    @staticmethod
    def dummy_interval(payload=None):
        return Interval3D((1,10), payload=payload)

    @staticmethod
    def query_that_throws_at_0(vids):
        output = []
        for vid in vids:
            if vid == 0:
                raise RuntimeError()
            output.append(TestRuntime.dummy_interval(vid))
        return IntervalSet3D(output)

    def assertCollectionEq(self, c1,c2):
        map1 = c1.get_grouped_intervals()
        map2 = c2.get_grouped_intervals()
        self.assertEqual(map1.keys(), map2.keys())
        for key in map1:
            is1 = map1[key]
            is2 = map2[key]
            self.assertIntervalSetEq(is1,is2)

    def assertIntervalSetEq(self, is1, is2):
        list1 = [(i.t, i.x, i.y, i.payload) for i in is1.get_intervals()]
        list2 = [(i.t, i.x, i.y, i.payload) for i in is2.get_intervals()]
        self.assertEqual(list1, list2)


    def test_single_process_runtime(self):
        vids = list(range(1000))
        rt = Runtime.inline()

        self.assertCollectionEq(
                rt.run(TestRuntime.query, vids)[0],
                TestRuntime.query(vids))

    def test_exception_inline(self):
        vids = list(range(2))
        rt = Runtime.inline()

        _, vids_with_err = rt.run(TestRuntime.query_that_throws_at_0, vids,
                print_error=False)
        self.assertEqual([0], vids_with_err)

    def test_forked_children(self):
        vids = list(range(10))
        rt = Runtime(get_forked_process_pool_factory())
        self.assertCollectionEq(
                rt.run(TestRuntime.query, vids, chunksize=3)[0],
                TestRuntime.query(vids))

    def test_forked_children_exception(self):
        vids = list(range(2))
        rt = Runtime(get_forked_process_pool_factory(1))
        _, vids_with_err = rt.run(TestRuntime.query_that_throws_at_0, vids,
                print_error=False)
        self.assertEqual([0], vids_with_err)

    def test_spawned_children(self):
        vids = list(range(10))
        rt = Runtime(get_spawned_process_pool_factory())
        self.assertCollectionEq(
                rt.run(TestRuntime.query, vids, chunksize=3)[0],
                TestRuntime.query(vids))

    def test_spawned_children_exception(self):
        vids = list(range(2))
        rt = Runtime(get_spawned_process_pool_factory())
        _, vids_with_err = rt.run(TestRuntime.query_that_throws_at_0, vids,
                print_error=False)
        self.assertEqual([0], vids_with_err)

    def test_returning_intervalset(self):
        vids = list(range(1,101))
        rt = Runtime(get_spawned_process_pool_factory())
        answer, _ = rt.run(TestRuntime.query_that_throws_at_0, vids)
        self.assertIntervalSetEq(answer,
                TestRuntime.query_that_throws_at_0(vids))

    def test_iterator(self):
        vids = list(range(1000))
        rt = Runtime(get_forked_process_pool_factory(5))
        gen = rt.get_result_iterator(TestRuntime.query, vids,
                randomize=False)
        for vid, result in zip(vids, gen):
            self.assertCollectionEq(result, TestRuntime.query([vid]))

    def test_inline_iterator(self):
        vids = list(range(1000))
        rt = Runtime.inline()
        gen = rt.get_result_iterator(TestRuntime.query, vids,
                randomize=True)
        for result in gen:
            self.assertCollectionEq(result, TestRuntime.query(result.keys()))
        
    def test_iterator_error(self):
        vids = list(range(2))
        rt = Runtime(get_spawned_process_pool_factory())
        gen = rt.get_result_iterator(TestRuntime.query_that_throws_at_0, vids,
                print_error=False)
        result = next(gen)
        self.assertIntervalSetEq(result,
                TestRuntime.query_that_throws_at_0([1]))
        with self.assertRaises(RekallRuntimeException):
            next(gen)

    def test_all_tasks_fail(self):
        vids = list(range(1))
        rt = Runtime.inline()
        with self.assertRaises(RekallRuntimeException):
            rt.run(TestRuntime.query_that_throws_at_0, vids,
                    print_error=False)






